# -*- coding: utf-8 -*-
import pytest

from flexget.api.app import base_message
from flexget.components.tmdb.api import ObjectsContainer as oc
from flexget.utils import json


@pytest.mark.online
class TestTMDBMovieLookupAPI:
    config = 'tasks: {}'

    def test_tmdb_movies_lookup_by_title(self, api_client, schema_match):
        # Bad API call
        rsp = api_client.get('/tmdb/movies/')
        assert rsp.status_code == 400, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(base_message, data)
        assert not errors

        rsp = api_client.get('/tmdb/movies/?title=the matrix/')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_object, data)
        assert not errors

        values = {'id': 603, 'name': 'The Matrix', 'year': 1999, 'imdb_id': 'tt0133093'}
        for field, value in values.items():
            assert data.get(field) == value

    def test_tmdb_movies_lookup_year_param(self, api_client, schema_match):
        rsp = api_client.get('/tmdb/movies/?title=the matrix reloaded&year=2003')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_object, data)
        assert not errors

        values = {'id': 604, 'name': 'The Matrix Reloaded', 'year': 2003, 'imdb_id': 'tt0234215'}
        for field, value in values.items():
            assert data.get(field) == value

    def test_tmdb_movies_lookup_posters_params(self, api_client, schema_match):
        rsp = api_client.get('/tmdb/movies/?title=the matrix&include_posters=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_object, data)
        assert not errors

        assert 'posters' in data
        assert len(data['posters']) > 0

    def test_tmdb_movies_lookup_backdrops_params(self, api_client, schema_match):
        rsp = api_client.get('/tmdb/movies/?title=the matrix&include_backdrops=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_object, data)
        assert not errors

        assert 'backdrops' in data
        assert len(data['backdrops']) > 0



import os
from json import loads, JSONDecodeError

from loguru import logger
from requests import Session
from requests.exceptions import RequestException

from flexget import plugin
from flexget.event import event
from flexget.utils.template import RenderError

logger = logger.bind(name='qbittorrent')


class OutputQBitTorrent:
    """
    Example:

      qbittorrent:
        username: <USERNAME> (default: (none))
        password: <PASSWORD> (default: (none))
        host: <HOSTNAME> (default: localhost)
        port: <PORT> (default: 8080)
        use_ssl: <SSL> (default: False)
        verify_cert: <VERIFY> (default: True)
        path: <OUTPUT_DIR> (default: (none))
        label: <LABEL> (default: (none))
        maxupspeed: <torrent upload speed limit> (default: 0)
        maxdownspeed: <torrent download speed limit> (default: 0)
        add_paused: <ADD_PAUSED> (default: False)
    """

    schema = {
        'anyOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'username': {'type': 'string'},
                    'password': {'type': 'string'},
                    'host': {'type': 'string'},
                    'port': {'type': 'integer'},
                    'use_ssl': {'type': 'boolean'},
                    'verify_cert': {'type': 'boolean'},
                    'path': {'type': 'string'},
                    'label': {'type': 'string'},
                    'maxupspeed': {'type': 'integer'},
                    'maxdownspeed': {'type': 'integer'},
                    'fail_html': {'type': 'boolean'},
                    'add_paused': {'type': 'boolean'},
                    'skip_check': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        ]
    }

    def __init__(self):
        super().__init__()
        self.session = Session()
        self.api_url_login = None
        self.api_url_upload = None
        self.api_url_download = None
        self.api_url_info = None
        self.url = None
        self.connected = False

    def _request(self, method, url, msg_on_fail=None, **kwargs):
        try:
            response = self.session.request(method, url, **kwargs)
            if response.text == "Ok.":
                return response
            msg = msg_on_fail if msg_on_fail else f'Failure. URL: {url}, data: {kwargs}'
        except RequestException as e:
            msg = str(e)
        raise plugin.PluginError(f'Error when trying to send request to qBittorrent: {msg}')

    def check_api_version(self, msg_on_fail, verify=True):
        try:
            url = self.url + "/api/v2/app/webapiVersion"
            response = self.session.request('get', url, verify=verify)
            if response.status_code != 404:
                self.api_url_login = '/api/v2/auth/login'
                self.api_url_upload = '/api/v2/torrents/add'
                self.api_url_download = '/api/v2/torrents/add'
                self.api_url_info = '/api/v2/torrents/info'
                return response

            url = self.url + "/version/api"
            response = self.session.request('get', url, verify=verify)
            if response.status_code != 404:
                self.api_url_login = '/login'
                self.api_url_upload = '/command/upload'
                self.api_url_download = '/command/download'
                self.api_url_info = '/query/torrents'
                return response

            msg = 'Failure. URL: {}'.format(url) if not msg_on_fail else msg_on_fail
        except RequestException as e:
            msg = str(e)
        raise plugin.PluginError(
            'Error when trying to send request to qBittorrent: {}'.format(msg)
        )

    def connect(self, config):
        """
        Connect to qBittorrent Web UI. Username and password not necessary
        if 'Bypass authentication for localhost' is checked and host is
        'localhost'.
        """
        self.url = '{}://{}:{}'.format(
            'https' if config['use_ssl'] else 'http', config['host'], config['port']
        )
        self.check_api_version('Check API version failed.', verify=config['verify_cert'])
        if config.get('username') and config.get('password'):
            data = {'username': config['username'], 'password': config['password']}
            self._request(
                'post',
                self.url + self.api_url_login,
                data=data,
                msg_on_fail='Authentication failed.',
                verify=config['verify_cert'],
            )
        logger.debug('Successfully connected to qBittorrent')
        self.connected = True

    def check_torrent_exists(self, hash_torrent, verify_cert):
        if not self.connected:
            raise plugin.PluginError('Not connected.')

        if not isinstance(hash_torrent, str):
            logger.error('Error getting torrent info, invalid hash {}', hash_torrent)
            return False

        hash_torrent = hash_torrent.lower()

        url = f'{self.url}{self.api_url_info}'
        params = {'hashes': hash_torrent}

        try:
            respose = self.session.request(
                'get',
                url,
                params=params,
                verify=verify_cert,
            )
        except RequestException as e:
            logger.error('Error getting torrent info, request to hash {} failed', hash_torrent)
            return False

        if respose.status_code != 200:
            logger.error(
                'Error getting torrent info, hash {} search returned',
                hash_torrent,
                respose.status_code,
            )
            return False

        check_file = respose.json()

        if isinstance(check_file, list) and check_file:
            logger.warning('File with hash {} already in qbittorrent', hash_torrent)
            return True

        return False

    def add_torrent_file(self, file_path, data, verify_cert):
        if not self.connected:
            raise plugin.PluginError('Not connected.')

        multipart_data = {k: (None, v) for k, v in data.items()}
        with open(file_path, 'rb') as f:
            multipart_data['torrents'] = f
            self._request(
                'post',
                self.url + self.api_url_upload,
                msg_on_fail='Failed to add file to qBittorrent',
                files=multipart_data,
                verify=verify_cert,
            )
        logger.debug('Added torrent file {} to qBittorrent', file_path)

    def add_torrent_url(self, url, data, verify_cert):
        if not self.connected:
            raise plugin.PluginError('Not connected.')

        data['urls'] = url
        multipart_data = {k: (None, v) for k, v in data.items()}
        self._request(
            'post',
            self.url + self.api_url_download,
            msg_on_fail='Failed to add file to qBittorrent',
            files=multipart_data,
            verify=verify_cert,
        )
        logger.debug('Added url {} to qBittorrent', url)

    @staticmethod
    def prepare_config(config):
        if isinstance(config, bool):
            config = {'enabled': config}
        config.setdefault('enabled', True)
        config.setdefault('host', 'localhost')
        config.setdefault('port', 8080)
        config.setdefault('use_ssl', False)
        config.setdefault('verify_cert', True)
        config.setdefault('label', '')
        config.setdefault('maxupspeed', 0)
        config.setdefault('maxdownspeed', 0)
        config.setdefault('fail_html', True)
        return config

    def add_entries(self, task, config):
        for entry in task.accepted:

            form_data = {}
            try:
                save_path = entry.render(entry.get('path', config.get('path', '')))
                if save_path:
                    form_data['savepath'] = save_path
            except RenderError as e:
                logger.error('Error setting path for {}: {}', entry['title'], e)

            label = entry.render(entry.get('label', config.get('label', '')))
            if label:
                form_data['label'] = label  # qBittorrent v3.3.3-
                form_data['category'] = label  # qBittorrent v3.3.4+

            add_paused = entry.get('add_paused', config.get('add_paused'))
            if add_paused:
                form_data['paused'] = 'true'

            skip_check = entry.get('skip_check', config.get('skip_check'))
            if skip_check:
                form_data['skip_checking'] = 'true'

            maxupspeed = entry.get('maxupspeed', config.get('maxupspeed'))
            if maxupspeed:
                form_data['upLimit'] = maxupspeed * 1024

            maxdownspeed = entry.get('maxdownspeed', config.get('maxdownspeed'))
            if maxdownspeed:
                form_data['dlLimit'] = maxdownspeed * 1024

            is_magnet = entry['url'].startswith('magnet:')

            if task.manager.options.test:
                logger.info('Test mode.')
                logger.info('Would add torrent to qBittorrent with:')
                if not is_magnet:
                    logger.info('File: {}', entry.get('file'))
                else:
                    logger.info('Url: {}', entry.get('url'))
                logger.info('Save path: {}', form_data.get('savepath'))
                logger.info('Label: {}', form_data.get('label'))
                logger.info('Paused: {}', form_data.get('paused', 'false'))
                logger.info('Skip Hash Check: {}', form_data.get('skip_checking', 'false'))
                if maxupspeed:
                    logger.info('Upload Speed Limit: {}', form_data.get('upLimit'))
                if maxdownspeed:
                    logger.info('Download Speed Limit: {}', form_data.get('dlLimit'))
                continue

            if self.check_torrent_exists(
                entry.get('torrent_info_hash'), config.get('verify_cert')
            ):
                continue

            if not is_magnet:
                if 'file' not in entry:
                    entry.fail('File missing?')
                    continue
                if not os.path.exists(entry['file']):
                    tmp_path = os.path.join(task.manager.config_base, 'temp')
                    logger.debug('entry: {}', entry)
                    logger.debug('temp: {}', ', '.join(os.listdir(tmp_path)))
                    entry.fail("Downloaded temp file '%s' doesn't exist!?" % entry['file'])
                    continue
                self.add_torrent_file(entry['file'], form_data, config['verify_cert'])
            else:
                self.add_torrent_url(entry['url'], form_data, config['verify_cert'])

    @plugin.priority(120)
    def on_task_download(self, task, config):
        """
        Call download plugin to generate torrent files to load into
        qBittorrent.
        """
        config = self.prepare_config(config)
        if not config['enabled']:
            return
        if 'download' not in task.config:
            download = plugin.get('download', self)
            download.get_temp_files(task, handle_magnets=True, fail_html=config['fail_html'])

    @plugin.priority(135)
    def on_task_output(self, task, config):
        """Add torrents to qBittorrent at exit."""
        if task.accepted:
            config = self.prepare_config(config)
            self.connect(config)
            self.add_entries(task, config)


@event('plugin.register')
def register_plugin():
    plugin.register(OutputQBitTorrent, 'qbittorrent', api_ver=2)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event

logger = logger.bind(name='metainfo_media_id')


class MetainfoMediaId(object):
    """
    Populate media_id field based on media type etc.
    """

    schema = {'type': 'boolean'}

    @plugin.priority(0)  # run after other metainfo plugins
    def on_task_metainfo(self, task, config):
        # Don't run if we are disabled
        if config is False:
            return
        for entry in task.entries:
            entry.add_lazy_fields(self.get_media_id, ['media_id'])

    @entry.register_lazy_lookup('media_id')
    def get_media_id(self, entry):
        # Try to generate a media id based on available parser fields
        media_id = None
        if entry.get('movie_name'):
            media_id = entry['movie_name']
            if entry.get('movie_year'):
                media_id = f'{media_id} {entry["movie_year"]}'
        elif entry.get('series_name'):
            media_id = entry['series_name']
            if entry.get('series_year'):
                media_id = f'{media_id} {entry["series_year"]}'

            if entry.get('series_episode'):
                media_id = f'{media_id} S{(entry.get("series_season") or 0):02}E{entry["series_episode"]:02}'
            elif entry.get('series_season'):
                media_id = f'{media_id} S{entry["series_season"]:02}E00'
            elif entry.get('series_date'):
                media_id = f'{media_id} {entry["series_date"]}'
            else:
                # We do not want a media id for series
                media_id = None

        if media_id:
            media_id = media_id.strip().lower()

        entry['media_id'] = media_id


@event('plugin.register')
def register_plugin():
    plugin.register(MetainfoMediaId, 'metainfo_media_id', api_ver=2, builtin=True)



import re

from loguru import logger

from flexget import plugin
from flexget.components.sites.urlrewriting import UrlRewritingError
from flexget.event import event
from flexget.utils.soup import get_soup

logger = logger.bind(name='FTDB')


class UrlRewriteFTDB:
    """FTDB RSS url_rewrite"""

    def url_rewritable(self, task, entry):
        # url = entry['url']
        if re.match(r'^http://www\.frenchtorrentdb\.com/[^/]+(?!/)[^/]+&rss=1', entry['url']):
            return True
        return False

    def url_rewrite(self, task, entry):
        old_url = entry['url']
        page_url = old_url.replace('DOWNLOAD', 'INFOS')
        page_url = page_url.replace('&rss=1', '')

        new_url = self.parse_download_page(page_url, task.requests)
        logger.debug('PAGE URL NEEDED : {}', page_url)
        logger.debug('{} OLD is rewrited to NEW {}', old_url, new_url)
        entry['url'] = new_url

    def parse_download_page(self, page_url, requests):
        page = requests.get(page_url)
        try:
            soup = get_soup(page.text)
        except Exception as e:
            raise UrlRewritingError(e)
        tag_a = soup.find("a", {"class": "dl_link"})
        if not tag_a:
            if soup.findAll(text="Connexion ?"):
                raise UrlRewritingError(
                    'You are not logged in,\
                                         check if your cookie for\
                                         authentication is up to date'
                )
            else:
                raise UrlRewritingError(
                    'You have reached your download\
                                        limit per 24hours, so I cannot\
                                        get the torrent'
                )
        torrent_url = "http://www.frenchtorrentdb.com" + tag_a.get('href') + "&js=1"
        logger.debug('TORRENT URL is : {}', torrent_url)
        return torrent_url


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteFTDB, 'frenchtorrentdb', interfaces=['urlrewriter'], api_ver=2)



TRANSLATE_MAP = {ord('&'): ' and '}
for char in '\'\\':
    TRANSLATE_MAP[ord(char)] = ''
for char in '_./-,[]():':
    TRANSLATE_MAP[ord(char)] = ' '


def normalize_series_name(name):
    """Returns a normalized version of the series name."""
    name = name.lower()
    name = name.replace('&amp;', ' and ')
    name = name.translate(TRANSLATE_MAP)  # Replaced some symbols with spaces
    name = ' '.join(name.split())
    return name



import os

from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.pathscrub import pathscrub
from flexget.utils.template import RenderError, render_from_entry

logger = logger.bind(name='symlink')


class Symlink:

    schema = {
        'oneOf': [
            {
                'title': 'specify options',
                'type': 'object',
                'properties': {
                    'to': {'type': 'string', 'format': 'path'},
                    'rename': {'type': 'string'},
                    'existing': {'type': 'string', 'enum': ['ignore', 'fail']},
                    'link_type': {'type': 'string', 'enum': ['soft', 'hard']},
                },
                'required': ['to'],
                'additionalProperties': False,
            },
            {'title': 'specify path', 'type': 'string', 'format': 'path'},
        ]
    }

    def prepare_config(self, config):
        if not isinstance(config, dict):
            config = {'to': config}

        config.setdefault('existing', 'fail')
        config.setdefault('link_type', 'soft')

        return config

    @plugin.priority(0)
    def on_task_output(self, task, config):
        if not config:
            return
        config = self.prepare_config(config)
        existing = config['existing']
        for entry in task.accepted:
            if 'location' not in entry:
                entry.fail('Does not have location field for symlinking')
                continue
            linkfrom = entry['location']
            linkfrom_path, linkfrom_name = os.path.split(linkfrom)

            # get the proper path and name in order of: entry, config, above split
            linkto_path = entry.get('link_to', config.get('to', linkfrom_path))
            if config.get('rename'):
                linkto_name = config['rename']
            elif entry.get('filename') and entry['filename'] != linkfrom_name:
                # entry specifies different filename than what was split from the path
                # since some inputs fill in filename it must be different in order to be used
                linkto_name = entry['filename']
            else:
                linkto_name = linkfrom_name

            try:
                linkto_path = entry.render(linkto_path)
            except RenderError as err:
                raise plugin.PluginError(
                    'Path value replacement `%s` failed: %s' % (linkto_path, err.args[0])
                )
            try:
                linkto_name = entry.render(linkto_name)
            except RenderError as err:
                raise plugin.PluginError(
                    'Filename value replacement `%s` failed: %s' % (linkto_name, err.args[0])
                )

            # Clean invalid characters with pathscrub plugin
            linkto_path = pathscrub(os.path.expanduser(linkto_path))
            linkto_name = pathscrub(linkto_name, filename=True)

            # Join path and filename
            linkto = os.path.join(linkto_path, linkto_name)
            if linkto == entry['location']:
                raise plugin.PluginWarning('source and destination are the same.')

            # Hardlinks for dirs will not be failed here
            if os.path.exists(linkto) and (
                config['link_type'] == 'soft' or os.path.isfile(linkfrom)
            ):
                msg = 'Symlink destination %s already exists' % linkto
                if existing == 'ignore':
                    logger.verbose(msg)
                else:
                    entry.fail(msg)
                continue
            logger.verbose('{}link `{}` to `{}`', config['link_type'], linkfrom, linkto)
            try:
                if config['link_type'] == 'soft':
                    os.symlink(linkfrom, linkto)
                else:
                    if os.path.isdir(linkfrom):
                        self.hard_link_dir(linkfrom, linkto, existing)
                    else:
                        dirname = os.path.dirname(linkto)
                        if not os.path.exists(dirname):
                            os.makedirs(dirname)
                        os.link(linkfrom, linkto)
            except OSError as e:
                entry.fail('Failed to create %slink, %s' % (config['link_type'], e))

    def hard_link_dir(self, path, destination, existing):
        if not os.path.exists(destination):
            try:
                os.makedirs(destination)
            except OSError as e:
                # Raised when it already exists, but are there other cases?
                logger.debug('Failed to create destination dir {}: {}', destination, e)
        # 'recursively' traverse and hard link
        working_dir = os.getcwd()
        os.chdir(path)  # change working dir to make dir joins easier
        for root, dirs, files in os.walk('.'):
            dst_dir = os.path.abspath(os.path.join(destination, root))
            for d in dirs:
                try:
                    os.mkdir(d)
                except OSError as e:
                    # Raised when it already exists, but are there other cases?
                    logger.debug('Failed to create subdir {}: {}', d, e)
            for f in files:
                src_file = os.path.join(root, f)
                dst_file = os.path.join(dst_dir, f)
                logger.debug('Hardlinking {} to {}', src_file, dst_file)
                try:
                    os.link(src_file, dst_file)
                except OSError as e:
                    logger.debug('Failed to create hardlink for file {}: {}', f, e)
                    if existing == 'fail':
                        raise  # reraise to fail the entry in the calling function

        os.chdir(working_dir)


@event('plugin.register')
def register_plugin():
    if os.name == 'nt':
        logger.trace('Symlinks not supported on Windows. Skipping Symlink plugin register.')
        return
    plugin.register(Symlink, 'symlink', api_ver=2)



import re
import sys
from datetime import datetime
from pathlib import Path

from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.entry import Entry
from flexget.event import event

logger = logger.bind(name='filesystem')


class Filesystem:
    """
    Uses local path content as an input. Can use recursion if configured.
    Recursion is False by default. Can be configured to true or get integer that will specify max depth in relation to
        base folder.
    All files/dir/symlinks are retrieved by default. Can be changed by using the 'retrieve' property.

    Example 1:: Single path

      filesystem: /storage/movies/

    Example 2:: List of paths

      filesystem:
         - /storage/movies/
         - /storage/tv/

    Example 3:: Object with list of paths

      filesystem:
        path:
          - /storage/movies/
          - /storage/tv/
        mask: '*.mkv'

    Example 4::

      filesystem:
        path:
          - /storage/movies/
          - /storage/tv/
        recursive: 4  # 4 levels deep from each base folder
        retrieve: files  # Only files will be retrieved

    Example 5::

      filesystem:
        path:
          - /storage/movies/
          - /storage/tv/
        recursive: yes  # No limit to depth, all sub dirs will be accessed
        retrieve:  # Only files and dirs will be retrieved
          - files
          - dirs

    """

    retrieval_options = ['files', 'dirs', 'symlinks']
    paths = one_or_more({'type': 'string', 'format': 'path'}, unique_items=True)

    schema = {
        'oneOf': [
            paths,
            {
                'type': 'object',
                'properties': {
                    'path': paths,
                    'mask': {'type': 'string'},
                    'regexp': {'type': 'string', 'format': 'regex'},
                    'recursive': {
                        'oneOf': [{'type': 'integer', 'minimum': 2}, {'type': 'boolean'}]
                    },
                    'retrieve': one_or_more(
                        {'type': 'string', 'enum': retrieval_options}, unique_items=True
                    ),
                },
                'required': ['path'],
                'additionalProperties': False,
            },
        ]
    }

    def prepare_config(self, config):
        from fnmatch import translate

        config = config

        # Converts config to a dict with a list of paths
        if not isinstance(config, dict):
            config = {'path': config}
        if not isinstance(config['path'], list):
            config['path'] = [config['path']]

        config.setdefault('recursive', False)
        # If mask was specified, turn it in to a regexp
        if config.get('mask'):
            config['regexp'] = translate(config['mask'])
        # If no mask or regexp specified, accept all files
        config.setdefault('regexp', '.')
        # Sets the default retrieval option to files
        config.setdefault('retrieve', self.retrieval_options)

        return config

    def create_entry(self, filepath: Path, test_mode):
        """
        Creates a single entry using a filepath and a type (file/dir)
        """
        filepath = filepath.absolute()
        entry = Entry()
        entry['location'] = str(filepath)
        entry['url'] = Path(filepath).absolute().as_uri()
        entry['filename'] = filepath.name
        if filepath.is_file():
            entry['title'] = filepath.stem
        else:
            entry['title'] = filepath.name
        file_stat = filepath.stat()
        try:
            entry['timestamp'] = datetime.fromtimestamp(file_stat.st_mtime)
        except Exception as e:
            logger.warning('Error setting timestamp for {}: {}', filepath, e)
            entry['timestamp'] = None
        entry['accessed'] = datetime.fromtimestamp(file_stat.st_atime)
        entry['modified'] = datetime.fromtimestamp(file_stat.st_mtime)
        entry['created'] = datetime.fromtimestamp(file_stat.st_ctime)
        if entry.isvalid():
            if test_mode:
                logger.info("Test mode. Entry includes:")
                logger.info(' Title: {}', entry['title'])
                logger.info(' URL: {}', entry['url'])
                logger.info(' Filename: {}', entry['filename'])
                logger.info(' Location: {}', entry['location'])
                logger.info(' Timestamp: {}', entry['timestamp'])
            return entry
        else:
            logger.error('Non valid entry created: {} ', entry)
            return

    def get_max_depth(self, recursion, base_depth):
        if recursion is False:
            return base_depth + 1
        elif recursion is True:
            return float('inf')
        else:
            return base_depth + recursion

    @staticmethod
    def get_folder_objects(folder: Path, recursion: bool):
        return folder.rglob('*') if recursion else folder.iterdir()

    def get_entries_from_path(
        self, path_list, match, recursion, test_mode, get_files, get_dirs, get_symlinks
    ):
        entries = []

        for folder in path_list:
            logger.verbose('Scanning folder {}. Recursion is set to {}.', folder, recursion)
            folder = Path(folder).expanduser()
            if not folder.exists():
                logger.error('{} does not exist (anymore.)', folder)
                continue
            logger.debug('Scanning {}', folder)
            base_depth = len(folder.parts)
            max_depth = self.get_max_depth(recursion, base_depth)
            folder_objects = self.get_folder_objects(folder, recursion)
            for path_object in folder_objects:
                logger.debug('Checking if {} qualifies to be added as an entry.', path_object)
                try:
                    path_object.exists()
                except UnicodeError:
                    logger.error(
                        'File {} not decodable with filesystem encoding: {}',
                        path_object,
                        sys.getfilesystemencoding(),
                    )
                    continue
                entry = None
                object_depth = len(path_object.parts)
                if object_depth <= max_depth:
                    if match(str(path_object)):
                        if (
                            (path_object.is_dir() and get_dirs)
                            or (path_object.is_symlink() and get_symlinks)
                            or (
                                path_object.is_file()
                                and not path_object.is_symlink()
                                and get_files
                            )
                        ):
                            entry = self.create_entry(path_object, test_mode)
                        else:
                            logger.debug(
                                "Path object's {} type doesn't match requested object types.",
                                path_object,
                            )
                        if entry and entry not in entries:
                            entries.append(entry)

        return entries

    def on_task_input(self, task, config):
        config = self.prepare_config(config)

        path_list = config['path']
        test_mode = task.options.test
        match = re.compile(config['regexp'], re.IGNORECASE).match
        recursive = config['recursive']
        get_files = 'files' in config['retrieve']
        get_dirs = 'dirs' in config['retrieve']
        get_symlinks = 'symlinks' in config['retrieve']

        logger.verbose('Starting to scan folders.')
        return self.get_entries_from_path(
            path_list, match, recursive, test_mode, get_files, get_dirs, get_symlinks
        )


@event('plugin.register')
def register_plugin():
    plugin.register(Filesystem, 'filesystem', api_ver=2)



import re

from dateutil.parser import parse as dateutil_parse
from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.requests import RequestException, TimedLimiter
from flexget.utils.soup import get_soup
from flexget.utils.tools import parse_filesize

logger = logger.bind(name='awesomehd')


class SearchAwesomeHD:
    """
    AwesomeHD search plugin.
    """

    schema = {
        'oneOf': [
            {
                'type': 'object',
                'properties': {
                    'passkey': {'type': 'string'},
                    'only_internal': {'type': 'boolean', 'default': False},
                },
                'required': ['passkey'],
                'additionalProperties': False,
            },
            {'type': 'string'},
        ]
    }

    base_url = 'https://awesome-hd.me/'

    def prepare_config(self, config):
        if not isinstance(config, dict):
            config = {'passkey': config, 'only_internal': False}

        return config

    @plugin.internet(logger)
    def search(self, task, entry, config):
        """
        Search for entries on AwesomeHD
        """
        # need lxml to parse xml
        try:
            import lxml  # noqa
        except ImportError as e:
            logger.debug('Error importing lxml: {}', e)
            raise plugin.DependencyError(
                'awesomehd', 'lxml', 'lxml module required. ImportError: %s' % e
            )

        config = self.prepare_config(config)

        # set a domain limit, but allow the user to overwrite it
        if 'awesome-hd.me' not in task.requests.domain_limiters:
            task.requests.add_domain_limiter(TimedLimiter('awesome-hd.me', '5 seconds'))

        entries = set()

        # Can only search for imdb
        if not entry.get('imdb_id'):
            logger.debug('Skipping entry {} because of missing imdb id', entry['title'])
            return entries

        # Standard search params
        params = {
            'passkey': config['passkey'],
            'internal': int(config['only_internal']),
            'action': 'imdbsearch',
            'imdb': entry['imdb_id'],
        }

        try:
            response = task.requests.get(self.base_url + 'searchapi.php', params=params).content

        except RequestException as e:
            logger.error('Failed to search for imdb id {}: {}', entry['imdb_id'], e)
            return entries

        try:
            soup = get_soup(response, 'xml')
            if soup.find('error'):
                logger.error(soup.find('error').get_text())
                return entries
        except Exception as e:
            logger.error('Failed to parse xml result for imdb id {}: {}', entry['imdb_id'], e)
            return entries

        authkey = soup.find('authkey').get_text()

        for result in soup.find_all('torrent'):
            # skip audio releases for now
            if not result.find('resolution').get_text():
                logger.debug('Skipping audio release')
                continue

            e = Entry()

            e['imdb_id'] = result.find('imdb').get_text()
            e['torrent_id'] = int(result.find('id').get_text())
            e['uploaded_at'] = dateutil_parse(result.find('time').get_text())
            e['content_size'] = parse_filesize('{} b'.format(result.find('size').get_text()))
            e['torrent_snatches'] = int(result.find('snatched').get_text())
            e['torrent_seeds'] = int(result.find('seeders').get_text())
            e['torrent_leeches'] = int(result.find('leechers').get_text())
            e['release_group'] = result.find('releasegroup').get_text()
            e['freeleech_percent'] = int((1 - float(result.find('freeleech').get_text())) * 100)
            e['encode_status'] = result.find('encodestatus').get_text()
            e['subtitles'] = result.find('subtitles').get_text().split(', ')

            e['url'] = (
                self.base_url
                + 'torrents.php?action=download&id={}&authkey={}&torrent_pass={}'.format(
                    e['torrent_id'], authkey, config['passkey']
                )
            )

            # Generate a somewhat sensible title
            audio = result.find('audioformat').get_text().replace('AC-3', 'AC3')  # normalize a bit
            source = result.find('media').get_text()
            encoder = result.find('encoding').get_text()
            # calling a WEB-DL a remux is pretty redundant
            if 'WEB' in source.upper():
                encoder = re.sub('REMUX', '', encoder, flags=re.IGNORECASE).strip()

            e[
                'title'
            ] = '{movie_name} {year} {resolution} {source} {audio} {encoder}-{release_group}'.format(
                movie_name=result.find('name').get_text(),
                year=result.find('year').get_text(),
                resolution=result.find('resolution').get_text(),
                source=source,
                audio=audio,
                encoder=encoder,
                release_group=e['release_group'],
            )

            entries.add(e)

        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(SearchAwesomeHD, 'awesomehd', interfaces=['search'], api_ver=2)



import os
import re

from loguru import logger

from flexget import plugin
from flexget.components.archives import utils as archiveutil
from flexget.event import event
from flexget.utils.template import RenderError, render_from_entry

logger = logger.bind(name='decompress')


def fail_entry_with_error(entry, error):
    """
    Log error message at error level and fail the entry
    """
    logger.error(error)
    entry.fail(error)


def open_archive_entry(entry):
    """
    Convenience function for opening archives from entries. Returns an archive.Archive object
    """
    archive_path = entry.get('location', '')
    if not archive_path:
        logger.error('Entry does not appear to represent a local file.')
        return
    if not os.path.exists(archive_path):
        logger.error('File no longer exists: {}', entry['location'])
        return
    try:
        archive = archiveutil.open_archive(archive_path)
    except archiveutil.BadArchive as error:
        fail_entry_with_error(entry, 'Bad archive: %s (%s)' % (archive_path, error))
    except archiveutil.NeedFirstVolume:
        logger.error('Not the first volume: {}', archive_path)
    except archiveutil.ArchiveError as error:
        fail_entry_with_error(entry, 'Failed to open Archive: %s (%s)' % (archive_path, error))
    else:
        return archive


def get_output_path(to, entry):
    """Determine which path to output to"""
    try:
        if to:
            return render_from_entry(to, entry)
        else:
            return os.path.dirname(entry.get('location'))
    except RenderError:
        raise plugin.PluginError('Could not render path: %s' % to)


def extract_info(info, archive, to, keep_dirs, test=False):
    """Extract ArchiveInfo object"""

    destination = get_destination_path(info, to, keep_dirs)

    if test:
        logger.info('Would extract: {} to {}', info.filename, destination)
        return
    logger.debug('Attempting to extract: {} to {}', info.filename, destination)
    try:
        info.extract(archive, destination)
    except archiveutil.FSError as error:
        logger.error('OS error while creating file: {} ({})', destination, error)
    except archiveutil.FileAlreadyExists as error:
        logger.warning('File already exists: {}', destination)
    except archiveutil.ArchiveError as error:
        logger.error('Failed to extract file: {} from {} ({})', info.filename, archive.path, error)


def get_destination_path(info, to, keep_dirs):
    """Generate the destination path for a given file"""

    path_suffix = info.path if keep_dirs else os.path.basename(info.path)

    return os.path.join(to, path_suffix)


def is_match(info, pattern):
    """Returns whether an info record matches the supplied regex"""
    match = re.compile(pattern, re.IGNORECASE).match
    is_match = bool(match(info.filename))

    if is_match:
        logger.debug('Found matching file: {}', info.filename)
    else:
        logger.debug('File did not match regexp: {}', info.filename)

    return is_match


class Decompress:
    r"""
    Extracts files from Zip or RAR archives. By default this plugin will extract to the same
    directory as the source archive, preserving directory structure from the archive.

    This plugin requires the rarfile Python module and unrar command line utility to extract RAR
    archives.

    Configuration:

    to:                 Destination path; supports Jinja2 templating on the input entry. Fields such
                        as series_name must be populated prior to input into this plugin using
                        metainfo_series or similar. If no path is specified, archive contents will
                        be extraced in the same directory as the archve itself.
    keep_dirs:          [yes|no] (default: yes) Indicates whether to preserve the directory
                        structure from within the archive in the destination path.
    mask:               Shell-style file mask; any matching files will be extracted. When used, this
                        field will override regexp.
    regexp:             Regular expression pattern; any matching files will be extracted. Overridden
                        by mask if specified.
    unrar_tool:         Specifies the path of the unrar tool. Only necessary if its location is not
                        defined in the operating system's PATH environment variable.
    delete_archive:     [yes|no] (default: no) Delete this archive after extraction is completed.


    Example:

      decompress:
        to: '/Volumes/External/TV/{{series_name}}/Season {{series_season}}/'
        keep_dirs: yes
        regexp: '.*s\d{1,2}e\d{1,2}.*\.mkv'
    """

    schema = {
        'anyOf': [
            {'type': 'boolean'},
            {
                'type': 'object',
                'properties': {
                    'to': {'type': 'string'},
                    'keep_dirs': {'type': 'boolean'},
                    'mask': {'type': 'string'},
                    'regexp': {'type': 'string', 'format': 'regex'},
                    'unrar_tool': {'type': 'string'},
                    'delete_archive': {'type': 'boolean'},
                },
                'additionalProperties': False,
            },
        ]
    }

    @staticmethod
    def prepare_config(config):
        """Prepare config for processing"""
        from fnmatch import translate

        if not isinstance(config, dict):
            config = {}

        config.setdefault('to', '')
        config.setdefault('keep_dirs', True)
        config.setdefault('unrar_tool', '')
        config.setdefault('delete_archive', False)

        # If mask was specified, turn it in to a regexp
        if 'mask' in config:
            config['regexp'] = translate(config['mask'])
        # If no mask or regexp specified, accept all files
        if 'regexp' not in config:
            config['regexp'] = '.'

        return config

    @staticmethod
    def handle_entry(entry, config, test=False):
        """
        Extract matching files into the directory specified

        Optionally delete the original archive if config.delete_archive is True
        """

        archive = open_archive_entry(entry)

        if not archive:
            return

        to = get_output_path(config['to'], entry)

        for info in archive.infolist():
            if is_match(info, config['regexp']):
                extract_info(info, archive, to, config['keep_dirs'], test=test)

        if config['delete_archive']:
            if not test:
                archive.delete()
            else:
                logger.info(f'Would delete archive {archive.path}')
                archive.close()
        else:
            archive.close()

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_output(self, task, config):
        """Task handler for archive_extract"""
        if isinstance(config, bool) and not config:
            return

        config = self.prepare_config(config)
        archiveutil.rarfile_set_tool_path(config)

        archiveutil.rarfile_set_path_sep(os.path.sep)

        for entry in task.accepted:
            self.handle_entry(entry, config, test=task.options.test)


@event('plugin.register')
def register_plugin():
    plugin.register(Decompress, 'decompress', api_ver=2)



import datetime
import math

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.cached_input import cached

from . import db

logger = logger.bind(name='trakt_calendar')

max_number_of_days = 31


class TraktCalendar:
    schema = {
        'type': 'object',
        'properties': {
            'start_day': {'type': 'integer', 'default': 0},
            'days': {'type': 'integer', 'default': 7},
            'account': {'type': 'string'},
            'strip_dates': {'type': 'boolean', 'default': False},
            'type': {'type': 'string', 'enum': ['shows', 'episodes']},
        },
        'required': ['type'],
        'additionalProperties': False,
    }

    # Series info
    series_map = {
        'trakt_series_name': 'title',
        'trakt_series_year': 'year',
        'imdb_id': 'ids.imdb',
        'tvdb_id': 'ids.tvdb',
        'tmdb_id': 'ids.tmdb',
        'trakt_show_id': 'ids.trakt',
        'trakt_slug': 'ids.slug',
        'tvrage_id': 'ids.tvrage',
        'trakt_trailer': 'trailer',
        'trakt_homepage': 'homepage',
        'trakt_series_runtime': 'runtime',
        'trakt_series_first_aired': 'first_aired',
        'trakt_series_air_time': 'airs.time',
        'trakt_series_air_day': 'airs.day',
        'trakt_series_air_timezone': 'airs.timezone',
        'trakt_series_content_rating': 'certification',
        'trakt_genres': 'genres',
        'trakt_series_network': 'network',
        'imdb_url': lambda s: s['ids']['imdb']
        and 'http://www.imdb.com/title/%s' % s['ids']['imdb'],
        'trakt_series_url': lambda s: s['ids']['slug']
        and 'https://trakt.tv/shows/%s' % s['ids']['slug'],
        'trakt_series_country': 'country',
        'trakt_series_status': 'status',
        'trakt_series_overview': 'overview',
        'trakt_series_rating': 'rating',
        'trakt_series_votes': 'votes',
        'trakt_series_language': 'language',
        'trakt_series_aired_episodes': 'aired_episodes',
        'trakt_languages': 'available_translations',
        'trakt_series_updated_at': 'updated_at',
    }

    # Episode info
    episode_map = {
        'trakt_ep_name': 'title',
        'trakt_ep_imdb_id': 'ids.imdb',
        'trakt_ep_tvdb_id': 'ids.tvdb',
        'trakt_ep_tmdb_id': 'ids.tmdb',
        'trakt_ep_tvrage': 'ids.tvrage',
        'trakt_episode_id': 'ids.trakt',
        'trakt_ep_first_aired': 'first_aired',
        'trakt_ep_overview': 'overview',
        'trakt_ep_abs_number': 'number_abs',
        'trakt_season': 'season',
        'trakt_episode': 'number',
        'trakt_ep_id': lambda ep: 'S%02dE%02d' % (ep['season'], ep['number']),
        'trakt_ep_languages': 'available_translations',
        'trakt_ep_runtime': 'runtime',
        'trakt_ep_updated_at': 'updated_at',
        'trakt_ep_rating': 'rating',
        'trakt_ep_votes': 'votes',
    }

    @cached('trakt_calendar', persist='2 hours')
    def on_task_input(self, task, config):
        start_date = datetime.datetime.now().date() + datetime.timedelta(days=config['start_day'])
        entries = set()

        # The API limit is max_number_of_days days for a single call.
        for start_day in range(0, config['days'], max_number_of_days):
            current_start_date = start_date + datetime.timedelta(days=start_day)
            number_of_days = min(config['days'] - start_day, max_number_of_days)

            logger.debug(
                'Start date for calendar: {}, end date: {}',
                current_start_date,
                current_start_date + datetime.timedelta(days=number_of_days),
            )

            url = db.get_api_url(
                'calendars',
                'my' if config.get('account') else 'all',
                'shows',
                current_start_date,
                number_of_days,
            )

            try:
                results = (
                    db.get_session(config.get('account'))
                    .get(url, params={'extended': 'full'})
                    .json()
                )
                logger.debug('Found {} calendar entries', len(results))
            except RequestException as e:
                raise plugin.PluginError('Error while fetching calendar: {0}'.format(e))

            for result in results:
                e = Entry()
                e.update_using_map(self.series_map, result['show'])
                if config['type'] == 'episodes':
                    e.update_using_map(self.episode_map, result['episode'])

                title = e['trakt_series_name']
                if not config['strip_dates']:
                    title = '{0} ({1})'.format(title, e['trakt_series_year'])

                url = e['trakt_series_url']

                if config['type'] == 'episodes':
                    title = '{0} S{1:02d}E{2:02d}'.format(
                        title, e['trakt_season'], e['trakt_episode']
                    )

                    url = '{0}/seasons/{1}/episodes/{2}'.format(
                        url, e['trakt_season'], e['trakt_episode']
                    )

                e['title'] = title
                e['url'] = url

                entries.add(e)

        return list(entries)


@event('plugin.register')
def register_plugin():
    plugin.register(TraktCalendar, 'trakt_calendar', api_ver=2, interfaces=['task'])



import os
import socket

import requests
from loguru import logger

from flexget import plugin
from flexget.event import event

try:
    import mechanicalsoup
except ImportError:
    mechanicalsoup = None


logger = logger.bind(name='formlogin')


class FormLogin:
    """
    Login on form
    """

    schema = {
        'type': 'object',
        'properties': {
            'url': {'type': 'string', 'format': 'url'},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'userfield': {'type': 'string'},
            'passfield': {'type': 'string'},
        },
        'required': ['url', 'username', 'password'],
        'additionalProperties': False,
    }

    def on_task_start(self, task, config):
        if not mechanicalsoup:
            raise plugin.PluginError(
                'mechanicalsoup required (python module), please install it.', logger
            )

        userfield = config.get('userfield', 'username')
        passfield = config.get('passfield', 'password')

        url = config['url']
        username = config['username']
        password = config['password']

        # Mechanicalsoup will override our session user agent header unless we explicitly pass it in
        user_agent = task.requests.headers.get('User-Agent')
        br = mechanicalsoup.StatefulBrowser(session=task.requests, user_agent=user_agent)

        try:
            response = br.open(url)
        except requests.RequestException:
            # TODO: improve error handling
            logger.opt(exception=True).debug('Exception getting login page.')
            raise plugin.PluginError('Unable to get login page', logger)

        # br.set_debug(True)

        num_forms = len(br.get_current_page().find_all('form'))
        if not num_forms:
            raise plugin.PluginError('Unable to find any forms on {}'.format(url), logger)
        try:
            for form_num in range(num_forms):
                br.select_form(nr=form_num)
                try:
                    br[userfield] = username
                    br[passfield] = password
                    break
                except mechanicalsoup.LinkNotFoundError:
                    pass
            else:
                received = os.path.join(task.manager.config_base, 'received')
                if not os.path.isdir(received):
                    os.mkdir(received)
                filename = os.path.join(received, '%s.formlogin.html' % task.name)
                with open(filename, 'wb') as f:
                    f.write(response.content)
                logger.critical(
                    'I have saved the login page content to {} for you to view', filename
                )
                raise plugin.PluginError('Unable to find login fields', logger)
        except socket.timeout:
            raise plugin.PluginError('Timed out on url %s' % url)

        try:
            br.submit_selected()
        except requests.RequestException:
            logger.opt(exception=True).debug('Exception submitting login form.')
            raise plugin.PluginError('Unable to post login form', logger)


@event('plugin.register')
def register_plugin():
    plugin.register(FormLogin, 'form', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

from . import db

logger = logger.bind(name='seen')


class FilterSeen:
    """
    Remembers previously downloaded content and rejects them in
    subsequent executions. Without this plugin FlexGet would
    download all matching content on every execution.

    This plugin is enabled on all tasks by default.
    See wiki for more information.
    """

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'string', 'enum': ['global', 'local']},
            {
                'type': 'object',
                'properties': {
                    'local': {'type': 'boolean'},
                    'fields': {
                        'type': 'array',
                        'items': {'type': 'string'},
                        "minItems": 1,
                        "uniqueItems": True,
                    },
                },
            },
        ]
    }

    def __init__(self):
        # remember and filter by these fields
        self.fields = ['title', 'url', 'original_url']
        self.keyword = 'seen'

    def prepare_config(self, config):
        if config is None:
            config = {}
        elif isinstance(config, bool):
            if config is False:
                return config
            else:
                config = {'local': False}
        elif isinstance(config, str):
            config = {'local': config == 'local'}

        config.setdefault('local', False)
        config.setdefault('fields', self.fields)
        return config

    @plugin.priority(plugin.PRIORITY_FIRST)
    def on_task_filter(self, task, config, remember_rejected=False):
        """Filter entries already accepted on previous runs."""
        config = self.prepare_config(config)
        if config is False:
            logger.debug('{} is disabled', self.keyword)
            return

        fields = config.get('fields')
        local = config.get('local')

        for entry in task.entries:
            # construct list of values looked
            values = []
            for field in fields:
                if field not in entry:
                    continue
                if entry[field] not in values and entry[field]:
                    values.append(str(entry[field]))
            if values:
                logger.trace('querying for: {}', ', '.join(values))
                # check if SeenField.value is any of the values
                found = db.search_by_field_values(
                    field_value_list=values, task_name=task.name, local=local, session=task.session
                )
                if found:
                    logger.debug(
                        "Rejecting '{}' '{}' because of seen '{}'",
                        entry['url'],
                        entry['title'],
                        found.value,
                    )
                    se = (
                        task.session.query(db.SeenEntry)
                        .filter(db.SeenEntry.id == found.seen_entry_id)
                        .one()
                    )
                    entry.reject(
                        'Entry with %s `%s` is already marked seen in the task %s at %s'
                        % (found.field, found.value, se.task, se.added.strftime('%Y-%m-%d %H:%M')),
                        remember=remember_rejected,
                    )

    def on_task_learn(self, task, config):
        """Remember succeeded entries"""
        config = self.prepare_config(config)
        if config is False:
            logger.debug('disabled')
            return

        fields = config.get('fields')
        local = config.get('local')

        if isinstance(config, list):
            fields.extend(config)

        for entry in task.accepted:
            self.learn(task, entry, fields=fields, local=local)
            # verbose if in learning mode
            if task.options.learn:
                logger.info("Learned '{}' (will skip this in the future)", entry['title'])

    def learn(self, task, entry, fields=None, reason=None, local=False):
        """Marks entry as seen"""
        # no explicit fields given, use default
        if not fields:
            fields = self.fields
        se = db.SeenEntry(entry['title'], str(task.name), reason, local)
        remembered = []
        for field in fields:
            if field not in entry:
                continue
            # removes duplicate values (eg. url, original_url are usually same)
            if entry[field] in remembered:
                continue
            remembered.append(entry[field])
            sf = db.SeenField(str(field), str(entry[field]))
            se.fields.append(sf)
            logger.debug("Learned '{}' (field: {}, local: {})", entry[field], field, local)
        # Only add the entry to the session if it has one of the required fields
        if se.fields:
            task.session.add(se)

    def forget(self, task, title):
        """Forget SeenEntry with :title:. Return True if forgotten."""
        se = task.session.query(db.SeenEntry).filter(db.SeenEntry.title == title).first()
        if se:
            logger.debug("Forgotten '{}' ({} fields)", title, len(se.fields))
            task.session.delete(se)
            return True


@event('plugin.register')
def register_plugin():
    plugin.register(FilterSeen, 'seen', builtin=True, api_ver=2)



"""Plugin for plex media server (www.plexapp.com)."""
import os
import re
from datetime import datetime
from os.path import basename
from socket import gethostbyname
from xml.dom.minidom import parseString

from loguru import logger

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import requests

logger = logger.bind(name='plex')


class InputPlex:
    """
    Uses a plex media server (www.plexapp.com) tv section as an input.

    'section'           Required parameter, numerical (/library/sections/<num>) or section name.
    'selection'         Can be set to different keys:
        - all                   : Default
        - unwatched             :
        - recentlyAdded         :
        - recentlyViewed        :
        - recentlyViewedShows   : Series only.
      'all' and 'recentlyViewedShows' will only produce a list of show names while the other three will produce
      filename and download url.
    'token'             Plex access token, used to connect to PMS
    'username'          Myplex (http://my.plexapp.com) username, used to connect to shared PMS'.
    'password'          Myplex (http://my.plexapp.com) password, used to connect to shared PMS'.
    'server'            Host/IP of PMS to connect to.
    'lowercase_title'   Convert filename (title) to lower case.
    'strip_non_alpha'   Sanitize filename (title), stripping all non-alphanumeric letters.
                        Better to turn off in case of non-english titles.
    'strip_year'        Remove year from title, ex: Show Name (2012) 01x01 => Show Name 01x01.
                        Movies will have year added to their filename unless this is set.
    'strip_parens'      Remove information in parens from title, ex: Show Name (UK)(2012) 01x01 => Show Name 01x01.
    'original_filename' Use filename stored in PMS instead of transformed name. lowercase_title and strip_year
                        will be ignored.
    'unwatched_only'    Request only unwatched media from PMS.
    'fetch'             What to download, can be set to the following values:
        - file          The file itself, default.
        - art           Series or movie art as configured in PMS
        - cover         Series cover for series, movie cover for movies.
        - thumb         Episode thumbnail, series only.
        - season_cover  Season cover, series only. If used in movies, movie cover will be set.


    Default parameters:
      server           : localhost
      port             : 32400
      selection        : all
      lowercase_title  : no
      strip_non_alpha  : yes
      strip_year       : yes
      strip_parens     : no
      original_filename: no
      unwatched_only   : no
      fetch            : file

    Example:

      plex:
        server: 192.168.1.23
        section: 3
        selection: recentlyAdded
        fetch: series_art
    """

    schema = {
        'type': 'object',
        'properties': {
            'server': {'type': 'string', 'default': '127.0.0.1'},
            'port': {'type': 'integer', 'default': 32400},
            'username': {'type': 'string'},
            'password': {'type': 'string'},
            'token': {'type': 'string'},
            'section': {'type': ['string', 'integer']},
            'selection': {'type': 'string', 'default': 'all'},
            'lowercase_title': {'type': 'boolean', 'default': False},
            'strip_non_alpha': {'type': 'boolean', 'default': True},
            'strip_year': {'type': 'boolean', 'default': True},
            'strip_parens': {'type': 'boolean', 'default': False},
            'original_filename': {'type': 'boolean', 'default': False},
            'unwatched_only': {'type': 'boolean', 'default': False},
            'fetch': {
                'type': 'string',
                'default': 'file',
                'enum': ['file', 'art', 'cover', 'thumb', 'season_cover'],
            },
        },
        'required': ['section'],
        'not': {
            'anyOf': [{'required': ['token', 'username']}, {'required': ['token', 'password']}]
        },
        'error_not': 'Cannot specify `username` and `password` with `token`',
        'dependencies': {'username': ['password'], 'password': ['username']},
        'additionalProperties': False,
    }

    def prepare_config(self, config):
        config['plexserver'] = config['server']
        config = self.plex_format_server(config)
        return config

    def plex_get_globalaccesstoken(self, config):
        header = {'X-Plex-Client-Identifier': 'flexget'}
        try:
            r = requests.post(
                'https://my.plexapp.com/users/sign_in.xml',
                auth=(config['username'], config['password']),
                headers=header,
            )
        except requests.RequestException as error:
            raise plugin.PluginError('Could not log in to myplex! Error: %s' % error)
        if 'Invalid email' in r.text:
            raise plugin.PluginError('Myplex: invalid username and/or password!')
        dom = parseString(r.text)
        globalaccesstoken = dom.getElementsByTagName('authentication-token')[
            0
        ].firstChild.nodeValue
        if not globalaccesstoken:
            raise plugin.PluginError('Myplex: could not find a server!')
        else:
            logger.debug('Myplex: Got global accesstoken: {}', globalaccesstoken)
        return globalaccesstoken

    def plex_get_accesstoken(self, config, globalaccesstoken=""):
        accesstoken = None
        if not globalaccesstoken:
            globalaccesstoken = self.plex_get_globalaccesstoken(config)
        if config['server'] in ('localhost', '127.0.0.1'):
            logger.debug('Server using localhost. Global Token will be used')
            return globalaccesstoken
        try:
            r = requests.get(
                "https://my.plexapp.com/pms/servers?X-Plex-Token=%s" % globalaccesstoken
            )
        except requests.RequestException as e:
            raise plugin.PluginError(
                "Could not get servers from my.plexapp.com using "
                "authentication-token: %s. (%s)" % (globalaccesstoken, e)
            )
        dom = parseString(r.text)
        for node in dom.getElementsByTagName('Server'):
            if config['server'] in (
                node.getAttribute('address'),
                node.getAttribute('localAddresses'),
            ):
                accesstoken = node.getAttribute('accessToken')
                logger.debug('Got plextoken: {}', accesstoken)
        if not accesstoken:
            raise plugin.PluginError('Could not retrieve accesstoken for %s.' % config['server'])
        else:
            return accesstoken

    def plex_format_server(self, config):
        if gethostbyname(config['server']) != config['server']:
            config['server'] = gethostbyname(config['server'])
        return config

    def plex_section_is_int(self, section):
        return isinstance(section, int)

    def on_task_input(self, task, config):
        config = self.prepare_config(config)
        urlconfig = {}
        urlappend = "?"
        entries = []
        if (
            config['unwatched_only']
            and config['section'] != 'recentlyViewedShows'
            and config['section'] != 'all'
        ):
            urlconfig['unwatched'] = '1'
        if config.get('token'):
            accesstoken = config['token']
            logger.debug('Using accesstoken: {}', accesstoken)
            urlconfig['X-Plex-Token'] = accesstoken
        elif config.get('username'):
            accesstoken = self.plex_get_accesstoken(config)
            logger.debug('Got accesstoken: {}', accesstoken)
            urlconfig['X-Plex-Token'] = accesstoken

        for key in urlconfig:
            urlappend += '%s=%s&' % (key, urlconfig[key])
        if not self.plex_section_is_int(config['section']):
            try:
                path = "/library/sections/"
                r = requests.get(
                    "http://%s:%d%s%s" % (config['plexserver'], config['port'], path, urlappend)
                )
            except requests.RequestException as e:
                raise plugin.PluginError('Error retrieving source: %s' % e)
            dom = parseString(r.text.encode("utf-8"))
            for node in dom.getElementsByTagName('Directory'):
                if node.getAttribute('title') == config['section']:
                    config['section'] = int(node.getAttribute('key'))
        if not self.plex_section_is_int(config['section']):
            raise plugin.PluginError('Could not find section \'%s\'' % config['section'])

        logger.debug(
            'Fetching http://{}:{}/library/sections/{}/{}{}',
            config['server'],
            config['port'],
            config['section'],
            config['selection'],
            urlappend,
        )
        try:
            path = "/library/sections/%s/%s" % (config['section'], config['selection'])
            r = requests.get(
                "http://%s:%d%s%s" % (config['plexserver'], config['port'], path, urlappend)
            )
        except requests.RequestException as e:
            raise plugin.PluginError(
                'There is no section with number %d. (%s)' % (config['section'], e)
            )
        dom = parseString(r.text.encode("utf-8"))
        plexsectionname = dom.getElementsByTagName('MediaContainer')[0].getAttribute('title1')
        viewgroup = dom.getElementsByTagName('MediaContainer')[0].getAttribute('viewGroup')

        logger.debug('Plex section "{}" is a "{}" section', plexsectionname, viewgroup)
        if viewgroup != "movie" and viewgroup != "show" and viewgroup != "episode":
            raise plugin.PluginError("Section is neither a movie nor tv show section!")
        domroot = "Directory"
        titletag = "title"
        if viewgroup == "episode":
            domroot = "Video"
            titletag = "grandparentTitle"
            thumbtag = "thumb"
            arttag = "art"
            seasoncovertag = "parentThumb"
            covertag = "grandparentThumb"
        elif viewgroup == "movie":
            domroot = "Video"
            titletag = "title"
            arttag = "art"
            seasoncovertag = "thumb"
            covertag = "thumb"
            if config['fetch'] == "thumb":
                raise plugin.PluginError(
                    "Movie sections does not have any thumbnails to download!"
                )
        for node in dom.getElementsByTagName(domroot):
            e = Entry()
            e['plex_server'] = config['plexserver']
            e['plex_port'] = config['port']
            e['plex_section'] = config['section']
            e['plex_section_name'] = plexsectionname
            e['plex_episode_thumb'] = ''

            title = node.getAttribute(titletag)
            if config['strip_year']:
                title = re.sub(r'^(.*)\(\d{4}\)(.*)', r'\1\2', title)
            if config['strip_parens']:
                title = re.sub(r'\(.*?\)', r'', title)
                title = title.strip()
            if config['strip_non_alpha']:
                title = re.sub(r'[\(\)]', r'', title)
                title = re.sub(r'&', r'And', title)
                title = re.sub(r'[^A-Za-z0-9- \']', r'', title)
            if config['lowercase_title']:
                title = title.lower()
            if viewgroup == "show":
                e['title'] = title
                e['url'] = 'NULL'
                entries.append(e)
                # show ends here.
                continue
            e['plex_art'] = "http://%s:%d%s%s" % (
                config['server'],
                config['port'],
                node.getAttribute(arttag),
                urlappend,
            )
            e['plex_cover'] = "http://%s:%d%s%s" % (
                config['server'],
                config['port'],
                node.getAttribute(covertag),
                urlappend,
            )
            e['plex_season_cover'] = "http://%s:%d%s%s" % (
                config['server'],
                config['port'],
                node.getAttribute(seasoncovertag),
                urlappend,
            )
            if viewgroup == "episode":
                e['plex_thumb'] = "http://%s:%d%s%s" % (
                    config['server'],
                    config['port'],
                    node.getAttribute('thumb'),
                    urlappend,
                )
                e['series_name'] = title
                e['plex_ep_name'] = node.getAttribute('title')
                season = int(node.getAttribute('parentIndex'))
                if node.getAttribute('parentIndex') == node.getAttribute('year'):
                    season = node.getAttribute('originallyAvailableAt')
                    filenamemap = "%s_%s%s_%s_%s_%s.%s"
                    episode = ""
                    e['series_id_type'] = 'date'
                    e['series_date'] = season
                elif node.getAttribute('index'):
                    episode = int(node.getAttribute('index'))
                    filenamemap = "%s_%02dx%02d_%s_%s_%s.%s"
                    e['series_season'] = season
                    e['series_episode'] = episode
                    e['series_id_type'] = 'ep'
                    e['series_id'] = 'S%02dE%02d' % (season, episode)
                else:
                    logger.debug(
                        "Could not get episode number for '{}' (Hint, ratingKey: {})",
                        title,
                        node.getAttribute('ratingKey'),
                    )
                    break
            elif viewgroup == "movie":
                filenamemap = "%s_%s_%s_%s.%s"

            e['plex_year'] = node.getAttribute('year')
            e['plex_added'] = datetime.fromtimestamp(int(node.getAttribute('addedAt')))
            e['plex_duration'] = node.getAttribute('duration')
            e['plex_summary'] = node.getAttribute('summary')
            e['plex_userrating'] = node.getAttribute('userrating')
            e['plex_key'] = node.getAttribute('ratingKey')
            count = node.getAttribute('viewCount')
            offset = node.getAttribute('viewOffset')
            if count:
                e['plex_status'] = "seen"
            elif offset:
                e['plex_status'] = "inprogress"
            else:
                e['plex_status'] = "unwatched"
            for media in node.getElementsByTagName('Media'):
                entry = Entry(e)
                vcodec = media.getAttribute('videoCodec')
                acodec = media.getAttribute('audioCodec')
                if media.hasAttribute('title'):
                    entry['plex_media_title'] = media.getAttribute('title')
                if media.hasAttribute('optimizedForStreaming'):
                    entry['plex_stream_optimized'] = media.getAttribute('optimizedForStreaming')
                if config['fetch'] == "file" or not config['fetch']:
                    container = media.getAttribute('container')
                else:
                    container = "jpg"
                resolution = media.getAttribute('videoResolution') + "p"
                for part in media.getElementsByTagName('Part'):
                    if config['fetch'] == "file" or not config['fetch']:
                        key = part.getAttribute('key')
                    elif config['fetch'] == "art":
                        key = node.getAttribute(arttag)
                    elif config['fetch'] == "cover":
                        key = node.getAttribute(arttag)
                    elif config['fetch'] == "season_cover":
                        key = node.getAttribute(seasoncovertag)
                    elif config['fetch'] == "thumb":
                        key = node.getAttribute(thumbtag)
                    # key = part.getAttribute('key')
                    duration = part.getAttribute('duration')
                    entry['plex_title'] = title
                    entry['title'] = title
                    if config['original_filename']:
                        filename, fileext = os.path.splitext(basename(part.getAttribute('file')))
                        if config['fetch'] != 'file':
                            filename += ".jpg"
                        else:
                            filename = "%s%s" % (filename, fileext)
                    else:
                        if viewgroup == "episode":
                            filename = filenamemap % (
                                title.replace(" ", "."),
                                season,
                                episode,
                                resolution,
                                vcodec,
                                acodec,
                                container,
                            )
                            entry['title'] = filename
                        elif viewgroup == "movie":
                            filename = filenamemap % (
                                title.replace(" ", "."),
                                resolution,
                                vcodec,
                                acodec,
                                container,
                            )
                            entry['title'] = filename
                    entry['plex_url'] = "http://%s:%d%s%s" % (
                        config['server'],
                        config['port'],
                        key,
                        urlappend,
                    )
                    entry['plex_path'] = key
                    entry['url'] = "http://%s:%d%s%s" % (
                        config['server'],
                        config['port'],
                        key,
                        urlappend,
                    )
                    entry['plex_duration'] = duration
                    entry['filename'] = filename
                    if key == "":
                        logger.debug("Could not find anything in PMS to download. Next!")
                    else:
                        entries.append(entry)
        return entries


@event('plugin.register')
def register_plugin():
    plugin.register(InputPlex, 'plex', api_ver=2)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.utils.database import with_session

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.thetvdb import api_tvdb as plugin_api_tvdb
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='api_tvdb')

logger = logger.bind(name='thetvdb_lookup')


class PluginThetvdbLookup:
    """Retrieves TheTVDB information for entries. Uses series_name,
    series_season, series_episode from series plugin.

    Example:

    thetvdb_lookup: yes

    Primarily used for passing thetvdb information to other plugins.
    Among these is the IMDB url for the series.

    This information is provided (via entry):
    series info:
      tvdb_series_name
      tvdb_rating
      tvdb_status (Continuing or Ended)
      tvdb_runtime (show runtime in minutes)
      tvdb_first_air_date
      tvdb_air_time
      tvdb_content_rating
      tvdb_genres
      tvdb_network
      tvdb_overview
      tvdb_banner
      tvdb_posters
      tvdb_airs_day_of_week
      tvdb_actors
      tvdb_language (en, fr, etc.)
      imdb_url (if available)
      zap2it_id (if available)
    episode info: (if episode is found)
      tvdb_ep_name
      tvdb_ep_overview
      tvdb_ep_directors
      tvdb_ep_writers
      tvdb_ep_air_date
      tvdb_ep_rating
      tvdb_ep_guest_stars
      tvdb_ep_image
    """

    # Series info
    series_map = {
        'tvdb_series_name': 'name',
        'tvdb_rating': 'rating',
        'tvdb_status': 'status',
        'tvdb_runtime': 'runtime',
        'tvdb_first_air_date': 'first_aired',
        'tvdb_air_time': 'airs_time',
        'tvdb_content_rating': 'content_rating',
        'tvdb_genres': lambda series: [g for g in series.genres],
        'tvdb_network': 'network',
        'tvdb_overview': 'overview',
        'tvdb_banner': 'banner',
        'tvdb_language': 'language',
        'tvdb_airs_day_of_week': 'airs_dayofweek',
        'imdb_url': lambda series: series.imdb_id
        and f'http://www.imdb.com/title/{series.imdb_id}',
        'imdb_id': 'imdb_id',
        'zap2it_id': 'zap2it_id',
        'tvdb_id': 'id',
        'tvdb_url': lambda series: f'http://thetvdb.com/index.php?tab=series&id={str(series.id)}',
    }

    series_actor_map = {'tvdb_actors': 'actors'}
    series_poster_map = {'tvdb_posters': 'posters'}

    # Episode info
    episode_map = {
        'tvdb_ep_name': 'name',
        'tvdb_ep_air_date': 'first_aired',
        'tvdb_ep_rating': 'rating',
        'tvdb_ep_image': 'image',
        'tvdb_ep_overview': 'overview',
        'tvdb_ep_directors': 'director',
        'tvdb_absolute_number': 'absolute_number',
        'tvdb_season': 'season_number',
        'tvdb_episode': 'episode_number',
        'tvdb_ep_id': lambda ep: f'S{ep.season_number:02d}E{ep.episode_number:02d}',
    }

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'object', 'properties': {'language': {'type': 'string', 'default': 'en'}}},
        ]
    }

    @with_session
    def series_lookup(self, entry, language, field_map, session=None):
        try:
            series = plugin_api_tvdb.lookup_series(
                entry.get('series_name', eval_lazy=False),
                tvdb_id=entry.get('tvdb_id', eval_lazy=False),
                language=entry.get('language', language),
                session=session,
            )
            entry.update_using_map(field_map, series)
        except LookupError as e:
            logger.debug(
                'Error looking up tvdb series information for {}: {}', entry['title'], e.args[0]
            )
        return entry

    @entry.register_lazy_lookup('tvdb_series_lookup')
    def lazy_series_lookup(self, entry, language):
        return self.series_lookup(entry, language, self.series_map)

    @entry.register_lazy_lookup('tvdb_series_actor_lookup')
    def lazy_series_actor_lookup(self, entry, language):
        return self.series_lookup(entry, language, self.series_actor_map)

    @entry.register_lazy_lookup('tvdb_series_poster_lookup')
    def lazy_series_poster_lookup(self, entry, language):
        return self.series_lookup(entry, language, self.series_poster_map)

    @entry.register_lazy_lookup('tvdb_episode_lookup')
    def lazy_episode_lookup(self, entry, language):
        try:
            season_offset = entry.get('thetvdb_lookup_season_offset', 0)
            episode_offset = entry.get('thetvdb_lookup_episode_offset', 0)
            if not isinstance(season_offset, int):
                logger.error('thetvdb_lookup_season_offset must be an integer')
                season_offset = 0
            if not isinstance(episode_offset, int):
                logger.error('thetvdb_lookup_episode_offset must be an integer')
                episode_offset = 0
            if season_offset != 0 or episode_offset != 0:
                logger.debug(
                    f'Using offset for tvdb lookup: season: {season_offset}, '
                    f'episode: {episode_offset}'
                )

            lookupargs = {
                'name': entry.get('series_name', eval_lazy=False),
                'tvdb_id': entry.get('tvdb_id', eval_lazy=False),
                'language': entry.get('language', language),
            }
            if entry['series_id_type'] == 'ep':
                lookupargs['season_number'] = entry['series_season'] + season_offset
                lookupargs['episode_number'] = entry['series_episode'] + episode_offset
            elif entry['series_id_type'] == 'sequence':
                lookupargs['absolute_number'] = entry['series_id'] + episode_offset
            elif entry['series_id_type'] == 'date':
                # TODO: Should thetvdb_lookup_episode_offset be used for date lookups as well?
                lookupargs['first_aired'] = entry['series_date']

            episode = plugin_api_tvdb.lookup_episode(**lookupargs)
            entry.update_using_map(self.episode_map, episode)
        except LookupError as e:
            logger.debug(
                'Error looking up tvdb episode information for {}: {}', entry['title'], e.args[0]
            )

    # Run after series and metainfo series
    @plugin.priority(110)
    def on_task_metainfo(self, task, config):
        if not config:
            return

        language = config['language'] if not isinstance(config, bool) else 'en'

        for entry in task.entries:
            # If there is information for a series lookup, register our series lazy fields
            if entry.get('series_name') or entry.get('tvdb_id', eval_lazy=False):
                entry.add_lazy_fields(
                    self.lazy_series_lookup, self.series_map, kwargs={'language': language}
                )
                entry.add_lazy_fields(
                    self.lazy_series_actor_lookup,
                    self.series_actor_map,
                    kwargs={'language': language},
                )
                entry.add_lazy_fields(
                    self.lazy_series_poster_lookup,
                    self.series_poster_map,
                    kwargs={'language': language},
                )

                # If there is season and ep info as well, register episode lazy fields
                if entry.get('series_id_type') in ('ep', 'sequence', 'date'):
                    if entry.get('season_pack'):
                        logger.verbose(
                            'TheTVDB API does not support season lookup at this time, skipping {}',
                            entry,
                        )
                    else:
                        entry.add_lazy_fields(
                            self.lazy_episode_lookup,
                            self.episode_map,
                            kwargs={'language': language},
                        )

    @property
    def series_identifier(self):
        """Returns the plugin main identifier type"""
        return 'tvdb_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        PluginThetvdbLookup, 'thetvdb_lookup', api_ver=2, interfaces=['task', 'series_metainfo']
    )



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.utils.tools import aggregate_inputs

logger = logger.bind(name='crossmatch')


class CrossMatch:
    """
    Perform action based on item on current task and other inputs.

    Example::

      crossmatch:
        from:
          - rss: http://example.com/
        fields:
          - title
        action: reject
        exact: yes
        case_sensitive: yes
    """

    schema = {
        'type': 'object',
        'properties': {
            'fields': {'type': 'array', 'items': {'type': 'string'}},
            'action': {'enum': ['accept', 'reject']},
            'from': {'type': 'array', 'items': {'$ref': '/schema/plugins?phase=input'}},
            'exact': {'type': 'boolean', 'default': True},
            'all_fields': {'type': 'boolean', 'default': False},
            'case_sensitive': {'type': 'boolean', 'default': True},
        },
        'required': ['fields', 'action', 'from'],
        'additionalProperties': False,
    }

    def on_task_filter(self, task, config):

        fields = config['fields']
        action = config['action']
        all_fields = config['all_fields']

        if not task.entries:
            logger.trace('Stopping crossmatch filter because of no entries to check')
            return

        match_entries = aggregate_inputs(task, config['from'])

        # perform action on intersecting entries
        for entry in task.entries:
            for generated_entry in match_entries:
                logger.trace('checking if {} matches {}', entry['title'], generated_entry['title'])
                common = self.entry_intersects(
                    entry,
                    generated_entry,
                    fields,
                    config.get('exact'),
                    config.get('case_sensitive'),
                )
                if common and (not all_fields or len(common) == len(fields)):
                    msg = 'intersects with %s on field(s) %s' % (
                        generated_entry['title'],
                        ', '.join(common),
                    )
                    for key in generated_entry:
                        if key not in entry:
                            entry[key] = generated_entry[key]
                    if action == 'reject':
                        entry.reject(msg)
                    if action == 'accept':
                        entry.accept(msg)

    def entry_intersects(self, e1, e2, fields=None, exact=True, case_sensitive=True):
        """
        :param e1: First :class:`flexget.entry.Entry`
        :param e2: Second :class:`flexget.entry.Entry`
        :param fields: List of fields which are checked
        :return: List of field names in common
        """

        if fields is None:
            fields = []

        common_fields = []

        for field in fields:
            # Doesn't really make sense to match if field is not in both entries
            if field not in e1 or field not in e2:
                logger.trace('field {} is not in both entries', field)
                continue

            if not case_sensitive and isinstance(e1[field], str):
                v1 = e1[field].lower()
            else:
                v1 = e1[field]
            if not case_sensitive and isinstance(e1[field], str):
                v2 = e2[field].lower()
            else:
                v2 = e2[field]

            try:
                if v1 == v2 or not exact and (v2 in v1 or v1 in v2):
                    common_fields.append(field)
                else:
                    logger.trace('not matching')
            except TypeError as e:
                # argument of type <type> is not iterable
                logger.trace('error matching fields: {}', str(e))

        return common_fields


@event('plugin.register')
def register_plugin():
    plugin.register(CrossMatch, 'crossmatch', api_ver=2)



import abc
import logging
import time

# Allow some request objects to be imported from here instead of requests
import warnings
from datetime import datetime, timedelta
from typing import Dict, Optional, Union, TYPE_CHECKING
from urllib.parse import urlparse
from urllib.request import urlopen

import requests
from loguru import logger
from requests import RequestException

from flexget import __version__ as version
from flexget.utils.tools import TimedDict, parse_timedelta

# If we use just 'requests' here, we'll get the logger created by requests, rather than our own
logger = logger.bind(name='utils.requests')

# Don't emit info level urllib3 log messages or below
logging.getLogger('requests.packages.urllib3').setLevel(logging.WARNING)
# same as above, but for systems where urllib3 isn't part of the requests pacakge (i.e., Ubuntu)
logging.getLogger('urllib3').setLevel(logging.WARNING)

# Time to wait before trying an unresponsive site again
WAIT_TIME = timedelta(seconds=60)
# Remembers sites that have timed out
unresponsive_hosts = TimedDict(WAIT_TIME)

if TYPE_CHECKING:
    from typing import TypedDict

    StateCacheDict = TypedDict(
        'StateCacheDict', {'tokens': Union[float, int], 'last_update': datetime}
    )


def is_unresponsive(url: str) -> bool:
    """
    Checks if host of given url has timed out within WAIT_TIME

    :param url: The url to check
    :return: True if the host has timed out within WAIT_TIME
    :rtype: bool
    """
    host = urlparse(url).hostname
    return host in unresponsive_hosts


def set_unresponsive(url: str) -> None:
    """
    Marks the host of a given url as unresponsive

    :param url: The url that timed out
    """
    host = urlparse(url).hostname
    if host in unresponsive_hosts:
        # If somehow this is called again before previous timer clears, don't refresh
        return
    unresponsive_hosts[host] = True


class DomainLimiter(abc.ABC):
    def __init__(self, domain: str) -> None:
        self.domain = domain

    @abc.abstractmethod
    def __call__(self) -> None:
        """This method will be called once before every request to the domain."""


class TokenBucketLimiter(DomainLimiter):
    """
    A token bucket rate limiter for domains.

    New instances for the same domain will restore previous values.
    """

    # This is just an in memory cache right now, it works for the daemon, and across tasks in a single execution
    # but not for multiple executions via cron. Do we need to store this to db?
    state_cache: Dict[str, 'StateCacheDict'] = {}

    def __init__(
        self,
        domain: str,
        tokens: Union[float, int],
        rate: Union[str, timedelta],
        wait: bool = True,
    ) -> None:
        """
        :param int tokens: Size of bucket
        :param rate: Amount of time to accrue 1 token. Either `timedelta` or interval string.
        :param bool wait: If true, will wait for a token to be available. If false, errors when token is not available.
        """
        super().__init__(domain)
        self.max_tokens = tokens
        self.rate = parse_timedelta(rate)
        self.wait = wait
        # Restore previous state for this domain, or establish new state cache
        self.state = self.state_cache.setdefault(
            domain, {'tokens': self.max_tokens, 'last_update': datetime.now()}
        )

    @property
    def tokens(self) -> Union[float, int]:
        return min(self.max_tokens, self.state['tokens'])

    @tokens.setter
    def tokens(self, value: Union[float, int]) -> None:
        self.state['tokens'] = value

    @property
    def last_update(self) -> datetime:
        return self.state['last_update']

    @last_update.setter
    def last_update(self, value: datetime) -> None:
        self.state['last_update'] = value

    def __call__(self) -> None:
        if self.tokens < self.max_tokens:
            regen = (datetime.now() - self.last_update).total_seconds() / self.rate.total_seconds()
            self.tokens += regen
        self.last_update = datetime.now()
        if self.tokens < 1:
            if not self.wait:
                raise RequestException('Requests to %s have exceeded their limit.' % self.domain)
            wait = self.rate.total_seconds() * (1 - self.tokens)
            # Don't spam console if wait is low
            if wait < 4:
                level = 'DEBUG'
            else:
                level = 'VERBOSE'
            logger.log(level, 'Waiting {:.2f} seconds until next request to {}', wait, self.domain)
            # Sleep until it is time for the next request
            time.sleep(wait)
        self.tokens -= 1


class TimedLimiter(TokenBucketLimiter):
    """Enforces a minimum interval between requests to a given domain."""

    def __init__(self, domain: str, interval: Union[str, timedelta]) -> None:
        super().__init__(domain, 1, interval)


def _wrap_urlopen(url: str, timeout: Optional[int] = None) -> requests.Response:
    """
    Handles alternate schemes using urllib, wraps the response in a requests.Response

    This is not installed as an adapter in requests, since urls without network locations
    (e.g. file:///somewhere) will cause errors

    """
    try:
        raw = urlopen(url, timeout=timeout)
    except OSError as e:
        msg = f'Error getting {url}: {e}'
        logger.error(msg)
        raise RequestException(msg)
    resp = requests.Response()
    resp.raw = raw
    # requests passes the `decode_content` kwarg to read
    orig_read = raw.read
    resp.raw.read = lambda size, **kwargs: orig_read(size)
    resp.status_code = raw.code or 200
    resp.headers = requests.structures.CaseInsensitiveDict(raw.headers)
    return resp


def limit_domains(url: str, limit_dict: Dict[str, DomainLimiter]) -> None:
    """
    If this url matches a domain in `limit_dict`, run the limiter.

    This is separated in to its own function so that limits can be disabled during unit tests with VCR.
    """
    for domain, limiter in limit_dict.items():
        if domain in url:
            limiter()
            break


class Session(requests.Session):
    """
    Subclass of requests Session class which defines some of our own defaults, records unresponsive sites,
    and raises errors by default.

    """

    def __init__(self, timeout: int = 30, max_retries: int = 1, **kwargs) -> None:
        """Set some defaults for our session if not explicitly defined."""
        super().__init__()
        self.timeout = timeout
        self.stream = True
        self.adapters['http://'].max_retries = max_retries
        # Stores min intervals between requests for certain sites
        self.domain_limiters: Dict[str, DomainLimiter] = {}
        self.headers.update({'User-Agent': 'FlexGet/%s (www.flexget.com)' % version})

    def add_cookiejar(self, cookiejar):
        """
        Merges cookies from `cookiejar` into cookiejar for this session.

        :param cookiejar: CookieJar instance to add to the session.
        """
        for cookie in cookiejar:
            self.cookies.set_cookie(cookie)

    def set_domain_delay(self, domain, delay):
        """
        DEPRECATED, use `add_domain_limiter`
        Registers a minimum interval between requests to `domain`

        :param domain: The domain to set the interval on
        :param delay: The amount of time between requests, can be a timedelta or string like '3 seconds'
        """
        warnings.warn(
            'set_domain_delay is deprecated, use add_domain_limiter',
            DeprecationWarning,
            stacklevel=2,
        )
        self.domain_limiters[domain] = TimedLimiter(domain, delay)

    def add_domain_limiter(self, limiter: DomainLimiter) -> None:
        """
        Add a limiter to throttle requests to a specific domain.

        :param DomainLimiter limiter: The `DomainLimiter` to add to the session.
        """
        self.domain_limiters[limiter.domain] = limiter

    def request(self, method: str, url: str, *args, **kwargs) -> requests.Response:
        """
        Does a request, but raises Timeout immediately if site is known to timeout, and records sites that timeout.
        Also raises errors getting the content by default.

        :param bool raise_status: If True, non-success status code responses will be raised as errors (True by default)
        :param disable_limiters: If True, any limiters configured for this session will be ignored for this request.
        """

        # Raise Timeout right away if site is known to timeout
        if is_unresponsive(url):
            raise requests.Timeout(
                f'Requests to this site ({urlparse(url).hostname}) have timed out recently. '
                'Waiting before trying again.'
            )

        # Run domain limiters for this url
        if not kwargs.pop('disable_limiters', False):
            limit_domains(url, self.domain_limiters)

        kwargs.setdefault('timeout', self.timeout)
        raise_status = kwargs.pop('raise_status', True)

        # If we do not have an adapter for this url, pass it off to urllib
        if not any(url.startswith(adapter) for adapter in self.adapters):
            logger.debug('No adaptor, passing off to urllib')
            return _wrap_urlopen(url, timeout=kwargs['timeout'])

        try:
            logger.debug(f'{method.upper()}ing URL {url} with args {args} and kwargs {kwargs}')
            result = super().request(method, url, *args, **kwargs)
        except requests.Timeout:
            # Mark this site in known unresponsive list
            set_unresponsive(url)
            raise

        if raise_status:
            result.raise_for_status()

        return result


# Define some module level functions that use our Session, so this module can be used like main requests module
def request(method: str, url: str, **kwargs) -> requests.Response:
    s = kwargs.pop('session', Session())
    return s.request(method=method, url=url, **kwargs)


def head(url: str, **kwargs) -> requests.Response:
    """Sends a HEAD request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param kwargs: Optional arguments that ``request`` takes.
    """
    kwargs.setdefault('allow_redirects', True)
    return request('head', url, **kwargs)


def get(url: str, **kwargs) -> requests.Response:
    """Sends a GET request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param kwargs: Optional arguments that ``request`` takes.
    """
    kwargs.setdefault('allow_redirects', True)
    return request('get', url, **kwargs)


def post(url: str, data=None, **kwargs) -> requests.Response:
    """Sends a POST request. Returns :class:`Response` object.

    :param url: URL for the new :class:`Request` object.
    :param data: (optional) Dictionary or bytes to send in the body of the :class:`Request`.
    :param kwargs: Optional arguments that ``request`` takes.
    """
    return request('post', url, data=data, **kwargs)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='priority')


# TODO: 1.2 figure out replacement for this
# Currently the manager reads this value directly out of the config when the 'execute' command is run, and this plugin
# does nothing but make the config key valid.
# In daemon mode, schedules should be made which run tasks in the proper order instead of using this.
class TaskPriority:
    """Set task priorities"""

    schema = {'type': 'integer'}

    def on_task_start(self, task, config):
        pass


@event('plugin.register')
def register_plugin():
    plugin.register(TaskPriority, 'priority', api_ver=2)



import re

from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event

from . import db

logger = logger.bind(name='next_trakt_episodes')


class NextTraktEpisodes:
    """
    Creates an entry for the latest or the next item in your watched or collected
    episodes in your trakt account.

    Syntax:

    next_trakt_episodes:
      account: <value>
      username: <value>
      position: <last|next>
      context: <collect|collected|watch|watched>
      list: <value>

    `account` option is required if the profile is private. `username` will default to account owner if `account` is
    specified.

    """

    schema = {
        'type': 'object',
        'properties': {
            'username': {'type': 'string'},
            'account': {'type': 'string'},
            'position': {'type': 'string', 'enum': ['last', 'next'], 'default': 'next'},
            'context': {
                'type': 'string',
                'enum': ['watched', 'collected', 'aired'],
                'default': 'watched',
            },
            'list': {'type': 'string'},
            'strip_dates': {'type': 'boolean', 'default': False},
        },
        'required': ['list'],
        'anyOf': [{'required': ['username']}, {'required': ['account']}],
        'error_anyOf': 'At least one of `username` or `account` options are needed.',
        'additionalProperties': False,
    }

    def on_task_input(self, task, config):
        if config.get('account') and not config.get('username'):
            config['username'] = 'me'
        session = db.get_session(account=config.get('account'))
        listed_series = {}
        args = ('users', config['username'])
        if config['list'] in ['collection', 'watchlist', 'watched']:
            args += (config['list'], 'shows')
        else:
            args += ('lists', db.make_list_slug(config['list']), 'items')
        try:
            data = session.get(db.get_api_url(args)).json()
        except RequestException as e:
            raise plugin.PluginError('Unable to get trakt list `%s`: %s' % (config['list'], e))
        if not data:
            logger.warning('The list "{}" is empty.', config['list'])
            return
        for item in data:
            if item.get('show'):
                if not item['show']['title']:
                    # Seems we can get entries with a blank show title sometimes
                    logger.warning('Found trakt list show without series name.')
                    continue
                ids = item['show']['ids']
                trakt_id = ids['trakt']
                listed_series[trakt_id] = {
                    'series_name': '%s (%s)' % (item['show']['title'], item['show']['year']),
                    'trakt_id': trakt_id,
                    'trakt_series_name': item['show']['title'],
                    'trakt_series_year': item['show']['year'],
                    'trakt_list': config.get('list'),
                }
                for id_name, id_value in ids.items():
                    entry_field_name = (
                        'trakt_slug' if id_name == 'slug' else id_name
                    )  # rename slug to trakt_slug
                    listed_series[trakt_id][entry_field_name] = id_value
        context = 'collection' if config['context'] == 'collected' else config['context']
        entries = []
        for trakt_id, fields in listed_series.items():
            if context == 'aired':
                url = db.get_api_url('shows', trakt_id, '{}_episode'.format(config['position']))
            else:
                url = db.get_api_url('shows', trakt_id, 'progress', context)
            try:
                response = session.get(url)
                if response.status_code == 204:
                    logger.debug('No {} episode for {}', config['position'], fields['series_name'])
                    continue
                data = response.json()
            except RequestException as e:
                raise plugin.PluginError(
                    'An error has occurred looking up: Trakt_id: %s Error: %s' % (trakt_id, e)
                )
            if context == 'aired':
                season_number = data['season']
                episode_number = data['number']
            elif config['position'] == 'next' and data.get('next_episode'):
                # If the next episode is already in the trakt database, we'll get it here
                season_number = data['next_episode']['season']
                episode_number = data['next_episode']['number']
            else:
                # If we need last ep, or next_episode was not provided, search for last ep
                for season in reversed(data['seasons']):
                    # Find the first season with collected/watched episodes
                    if not season['completed']:
                        continue
                    season_number = season['number']
                    # Pick the highest collected/watched episode
                    episode_number = max(
                        item['number'] for item in season['episodes'] if item['completed']
                    )
                    # If we are in next episode mode, we have to increment this number
                    if config['position'] == 'next':
                        if season['completed'] >= season['aired']:
                            # TODO: next_episode doesn't count unaired episodes right now, this will skip to next
                            # season too early when there are episodes left to air this season.
                            season_number += 1
                        episode_number = 1
                    break
                else:
                    if config['position'] != 'next':
                        # There were no watched/collected episodes, nothing to emit in 'last' mode
                        continue
                    season_number = episode_number = 1

            if season_number and episode_number:
                if config.get('strip_dates'):
                    # remove year from end of series_name if present
                    fields['series_name'] = re.sub(r'\s+\(\d{4}\)$', '', fields['series_name'])
                entry = self.make_entry(fields, season_number, episode_number)
                entries.append(entry)
        return entries

    @staticmethod
    def make_entry(fields, season, episode):
        entry = Entry()
        entry.update(fields)
        entry['series_season'] = season
        entry['series_episode'] = episode
        entry['series_id_type'] = 'ep'
        entry['series_id'] = 'S%02dE%02d' % (season, episode)
        entry['title'] = entry['series_name'] + ' ' + entry['series_id']
        entry['url'] = 'https://trakt.tv/shows/%s/seasons/%s/episodes/%s' % (
            fields['trakt_id'],
            season,
            episode,
        )
        return entry


@event('plugin.register')
def register_plugin():
    plugin.register(NextTraktEpisodes, 'next_trakt_episodes', api_ver=2)



from loguru import logger

from flexget import entry, plugin
from flexget.event import event
from flexget.manager import Session
from flexget.utils.log import log_once

try:
    # NOTE: Importing other plugins is discouraged!
    from flexget.components.imdb.utils import extract_id
except ImportError:
    raise plugin.DependencyError(issued_by=__name__, missing='imdb')


logger = logger.bind(name='tmdb_lookup')


class PluginTmdbLookup:
    """Retrieves tmdb information for entries.

    Example:
        tmdb_lookup: yes
    """

    field_map = {
        'tmdb_name': 'name',
        'tmdb_id': 'id',
        'imdb_id': 'imdb_id',
        'tmdb_year': 'year',
        'tmdb_popularity': 'popularity',
        'tmdb_rating': 'rating',
        # Make sure we don't stor the db association proxy directly on the entry
        'tmdb_genres': lambda movie: list(movie.genres),
        'tmdb_released': 'released',
        'tmdb_votes': 'votes',
        # Just grab the top 5 posters
        'tmdb_posters': lambda movie: [poster.url('original') for poster in movie.posters[:5]],
        'tmdb_backdrops': lambda movie: [
            backdrop.url('original') for backdrop in movie.backdrops[:5]
        ],
        'tmdb_runtime': 'runtime',
        'tmdb_tagline': 'tagline',
        'tmdb_budget': 'budget',
        'tmdb_revenue': 'revenue',
        'tmdb_homepage': 'homepage',
        # Generic fields filled by all movie lookup plugins:
        'movie_name': 'name',
        'movie_year': 'year',
    }

    schema = {
        'oneOf': [
            {'type': 'boolean'},
            {'type': 'object', 'properties': {'language': {'type': 'string', 'default': 'en'}}},
        ]
    }

    @entry.register_lazy_lookup('tmdb_lookup')
    def lazy_loader(self, entry, language):
        """Does the lookup for this entry and populates the entry fields."""
        lookup = plugin.get('api_tmdb', self).lookup

        imdb_id = entry.get('imdb_id', eval_lazy=False) or extract_id(
            entry.get('imdb_url', eval_lazy=False)
        )
        try:
            with Session() as session:
                movie = lookup(
                    smart_match=entry['title'],
                    tmdb_id=entry.get('tmdb_id', eval_lazy=False),
                    imdb_id=imdb_id,
                    language=language,
                    session=session,
                )
                entry.update_using_map(self.field_map, movie)
        except LookupError:
            log_once('TMDB lookup failed for %s' % entry['title'], logger, 'WARNING')

    def lookup(self, entry, language):
        """
        Populates all lazy fields to an Entry. May be called by other plugins
        requiring tmdb info on an Entry

        :param entry: Entry instance
        """
        entry.add_lazy_fields(self.lazy_loader, self.field_map, kwargs={'language': language})

    def on_task_metainfo(self, task, config):
        if not config:
            return
        language = config['language'] if not isinstance(config, bool) else 'en'

        for entry in task.entries:
            self.lookup(entry, language)

    @property
    def movie_identifier(self):
        """Returns the plugin main identifier type"""
        return 'tmdb_id'


@event('plugin.register')
def register_plugin():
    plugin.register(
        PluginTmdbLookup, 'tmdb_lookup', api_ver=2, interfaces=['task', 'movie_metainfo']
    )



from flexget.api.app import base_message
from flexget.components.history.api import ObjectsContainer as OC
from flexget.components.history.db import History
from flexget.manager import Session
from flexget.utils import json


class TestHistoryAPI:
    config = "{'tasks': {}}"

    def test_history(self, api_client, schema_match):
        rsp = api_client.get('/history/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data == []

        history_entry = dict(
            task='test_task1',
            title='test_title1',
            url='test_url1',
            filename='test_filename1',
            details='test_details1',
        )

        with Session() as session:
            item = History()
            for key, value in history_entry.items():
                setattr(item, key, value)
            session.add(item)
            session.commit()

        rsp = api_client.get('/history/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        for key, value in history_entry.items():
            assert data[0][key] == value

        rsp = api_client.get('/history/?task=test_task1')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        for key, value in history_entry.items():
            assert data[0][key] == value

        rsp = api_client.get('/history/?task=bla')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data == []


class TestHistoryPaginationAPI:
    config = "{'tasks': {}}"

    def test_history_pagination(self, api_client, schema_match, link_headers):
        history_entry = dict(
            task='test_task_',
            title='test_title_',
            url='test_url_',
            filename='test_filename_',
            details='test_details_',
        )
        num_of_entries = 200

        with Session() as session:
            for i in range(num_of_entries):
                item = History()
                for key, value in history_entry.items():
                    setattr(item, key, value + str(i))
                session.add(item)

        rsp = api_client.get('/history/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert len(data) == 50  # Default page size
        assert int(rsp.headers['total-count']) == 200
        assert int(rsp.headers['count']) == 50

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 2

        rsp = api_client.get('/history/?per_page=100')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert len(data) == 100
        assert int(rsp.headers['count']) == 100

        links = link_headers(rsp)
        assert links['last']['page'] == 2
        assert links['next']['page'] == 2

        # Per page is limited to 100
        rsp = api_client.get('/history/?per_page=200')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert len(data) == 100

        rsp = api_client.get('/history/?page=2&sort_by=id&order=asc')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data[0]['task'] == 'test_task_50'

        links = link_headers(rsp)
        assert links['last']['page'] == 4
        assert links['next']['page'] == 3
        assert links['prev']['page'] == 1

        # Non existent page
        rsp = api_client.get('/history/?page=5')
        assert rsp.status_code == 404
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(base_message, data)
        assert not errors

    def test_history_sorting(self, api_client, schema_match, link_headers):
        history_entry1 = dict(
            task='test_task_1',
            title='test_title_a',
            url='test_url_1',
            filename='test_filename_a',
            details='test_details_1',
        )

        history_entry2 = dict(
            task='test_task_2',
            title='test_title_b',
            url='test_url_2',
            filename='test_filename_b',
            details='test_details_2',
        )

        history_entry3 = dict(
            task='test_task_3',
            title='test_title_c',
            url='test_url_3',
            filename='test_filename_c',
            details='test_details_3',
        )

        entries = [history_entry1, history_entry2, history_entry3]

        with Session() as session:
            for entry in entries:
                item = History()
                for key, value in entry.items():
                    setattr(item, key, value)
                session.add(item)

        rsp = api_client.get('/history/?sort_by=id')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data[0]['id'] == 3
        assert int(rsp.headers['total-count']) == 3
        assert int(rsp.headers['count']) == 3

        rsp = api_client.get('/history/?sort_by=task&order=asc')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data[0]['task'] == 'test_task_1'

        rsp = api_client.get('/history/?sort_by=details')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert data[0]['details'] == 'test_details_3'

        rsp = api_client.get('/history/?per_page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert len(data) == 2
        assert int(rsp.headers['total-count']) == 3
        assert int(rsp.headers['count']) == 2

        rsp = api_client.get('/history/?per_page=2&page=2')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.history_list_object, data)
        assert not errors

        assert len(data) == 1
        assert int(rsp.headers['total-count']) == 3
        assert int(rsp.headers['count']) == 1



from urllib.parse import quote

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='animeindex')


# http://tracker.anime-index.org/index.php?page=torrent-details&id=b8327fdf9003e87446c8b3601951a9a65526abb2
# http://tracker.anime-index.org/download.php?id=b8327fdf9003e87446c8b3601951a9a65526abb2&f=[DeadFish]%20Yowamushi%20Pedal:%20Grande%20Road%20-%2002%20[720p][AAC].mp4.torrent


class UrlRewriteAnimeIndex:
    """AnimeIndex urlrewriter."""

    def url_rewritable(self, task, entry):
        return entry['url'].startswith(
            'http://tracker.anime-index.org/index.php?page=torrent-details&id='
        )

    def url_rewrite(self, task, entry):
        entry['url'] = entry['url'].replace('index.php?page=torrent-details&', 'download.php?')
        entry['url'] += '&f=%s.torrent' % (quote(entry['title'], safe=''))


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteAnimeIndex, 'animeindex', interfaces=['urlrewriter'], api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='details')


class PluginDetails:
    def on_task_start(self, task, config):
        # Make a flag for tasks to declare if it is ok not to produce entries
        task.no_entries_ok = False

    @plugin.priority(-512)
    def on_task_input(self, task, config):
        if not task.entries:
            if task.no_entries_ok:
                logger.verbose('Task didn\'t produce any entries.')
            else:
                logger.warning(
                    "Task didn't produce any entries. "
                    "This is likely due to a mis-configured or non-functional input."
                )
        else:
            logger.verbose('Produced {} entries.', len(task.entries))

    @plugin.priority(-512)
    def on_task_download(self, task, config):
        # Needs to happen as the first in download, so it runs after urlrewrites
        # and IMDB queue acceptance.
        logger.verbose(
            'Summary - Accepted: {} (Rejected: {} Undecided: {} Failed: {})',
            len(task.accepted),
            len(task.rejected),
            len(task.undecided),
            len(task.failed),
        )


class NoEntriesOk:
    """Allows manually silencing the warning message for tasks that regularly produce no entries."""

    schema = {'type': 'boolean'}

    # Run after details plugin task_start
    @plugin.priority(127)
    def on_task_start(self, task, config):
        task.no_entries_ok = config


@event('plugin.register')
def register_plugin():
    plugin.register(PluginDetails, 'details', builtin=True, api_ver=2)
    plugin.register(NoEntriesOk, 'no_entries_ok', api_ver=2)



from urllib.parse import quote

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='cinemageddon')


class UrlRewriteCinemageddon:
    """Cinemageddon urlrewriter."""

    def url_rewritable(self, task, entry):
        return entry['url'].startswith('http://cinemageddon.net/details.php?id=')

    def url_rewrite(self, task, entry):
        entry['url'] = entry['url'].replace('details.php?id=', 'download.php?id=')
        entry['url'] += '&name=%s.torrent' % (quote(entry['title'], safe=''))


@event('plugin.register')
def register_plugin():
    plugin.register(UrlRewriteCinemageddon, 'cinemageddon', interfaces=['urlrewriter'], api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='p_priority')


class PluginPriority:
    """
    Allows modifying plugin priorities from default values.

    Example:

    plugin_priority:
      ignore: 50
      series: 100
    """

    schema = {'type': 'object', 'additionalProperties': {'type': 'integer'}}

    def __init__(self):
        self.priorities = {}

    def on_task_start(self, task, config):
        self.priorities = {}
        names = []
        for name, priority in config.items():
            names.append(name)
            originals = self.priorities.setdefault(name, {})
            for phase, phase_event in plugin.plugins[name].phase_handlers.items():
                originals[phase] = phase_event.priority
                logger.debug('stored {} original value {}', phase, phase_event.priority)
                phase_event.priority = priority
                logger.debug('set {} new value {}', phase, priority)
        logger.debug('Changed priority for: {}', ', '.join(names))

    def on_task_exit(self, task, config):
        if not self.priorities:
            logger.debug('nothing changed, aborting restore')
            return
        names = []
        for name in list(config.keys()):
            names.append(name)
            originals = self.priorities[name]
            for phase, priority in originals.items():
                plugin.plugins[name].phase_handlers[phase].priority = priority
        logger.debug('Restored priority for: {}', ', '.join(names))
        self.priorities = {}

    on_task_abort = on_task_exit


@event('plugin.register')
def register_plugin():
    plugin.register(PluginPriority, 'plugin_priority', api_ver=2)



from datetime import datetime, timedelta

from loguru import logger
from sqlalchemy import Boolean, Column, DateTime, Integer, String, Unicode, select

from flexget import db_schema
from flexget.entry import Entry
from flexget.event import event
from flexget.utils import json, serialization
from flexget.utils.database import entry_synonym
from flexget.utils.sqlalchemy_utils import table_schema

logger = logger.bind(name='pending_approval')
Base = db_schema.versioned_base('pending_approval', 1)


@db_schema.upgrade('pending_approval')
def upgrade(ver, session):
    if ver == 0:
        table = table_schema('pending_entries', session)
        for row in session.execute(select([table.c.id, table.c.json])):
            if not row['json']:
                # Seems there could be invalid data somehow. See #2590
                continue
            data = json.loads(row['json'], decode_datetime=True)
            # If title looked like a date, make sure it's a string
            title = str(data.pop('title'))
            e = Entry(title=title, **data)
            session.execute(
                table.update().where(table.c.id == row['id']).values(json=serialization.dumps(e))
            )

        ver = 1
    return ver


class PendingEntry(Base):
    __tablename__ = 'pending_entries'

    id = Column(Integer, primary_key=True, autoincrement=True, nullable=False)
    task_name = Column(Unicode)
    title = Column(Unicode)
    url = Column(String)
    approved = Column(Boolean)
    _json = Column('json', Unicode)
    entry = entry_synonym('_json')
    added = Column(DateTime, default=datetime.now)

    def __init__(self, task_name, entry):
        self.task_name = task_name
        self.title = entry['title']
        self.url = entry['url']
        self.approved = False
        self.entry = entry

    def __repr__(self):
        return '<PendingEntry(task_name={},title={},url={},approved={})>'.format(
            self.task_name, self.title, self.url, self.approved
        )

    def to_dict(self):
        return {
            'id': self.id,
            'task_name': self.task_name,
            'title': self.title,
            'url': self.url,
            'approved': self.approved,
            'added': self.added,
        }


@event('manager.db_cleanup')
def db_cleanup(manager, session):
    # Clean unapproved entries older than 1 year
    deleted = (
        session.query(PendingEntry)
        .filter(PendingEntry.added < datetime.now() - timedelta(days=365))
        .delete()
    )
    if deleted:
        logger.info('Purged {} pending entries older than 1 year', deleted)


def list_pending_entries(
    session, task_name=None, approved=None, start=None, stop=None, sort_by='added', descending=True
):
    logger.debug('querying pending entries')
    query = session.query(PendingEntry)
    if task_name:
        query = query.filter(PendingEntry.task_name == task_name)
    if approved is not None:
        query = query.filter(PendingEntry.approved == approved)
    if descending:
        query = query.order_by(getattr(PendingEntry, sort_by).desc())
    else:
        query = query.order_by(getattr(PendingEntry, sort_by))
    return query.slice(start, stop).all()


def get_entry_by_id(session, entry_id):
    return session.query(PendingEntry).filter(PendingEntry.id == entry_id).one()



import json
from collections.abc import MutableSet
from urllib.parse import quote, urlparse

import requests
from loguru import logger
from requests import RequestException

from flexget import plugin
from flexget.entry import Entry
from flexget.event import event
from flexget.utils.qualities import Requirements

logger = logger.bind(name='radarr')


class RadarrRequestError(Exception):
    def __init__(self, value, logger=logger, **kwargs):
        super().__init__()
        # Value is expected to be a string
        value = str(value)
        self.value = value
        self.logger = logger
        self.kwargs = kwargs

    def __str__(self):
        return self.value


class RadarrMovieAlreadyExistsError(Exception):
    pass


def spec_exception_from_response_ex(radarr_request_ex):
    error_message = None

    if "error_message" in radarr_request_ex.kwargs:
        error_message = radarr_request_ex.kwargs["error_message"]

    if not error_message:
        return None

    if error_message.lower() == "this movie has already been added":
        return RadarrMovieAlreadyExistsError()


def request_get_json(url, headers):
    """ Makes a GET request and returns the JSON response """
    try:
        response = requests.get(url, headers=headers, timeout=10)  # TODO: HANGS HERE
        if response.status_code == 200:
            return response.json()
        else:
            raise RadarrRequestError(
                "Invalid response received from Radarr: %s" % response.content
            )
    except RequestException as e:
        raise RadarrRequestError("Unable to connect to Radarr at %s. Error: %s" % (url, e))


def request_delete_json(url, headers):
    """ Makes a DELETE request and returns the JSON response """
    try:
        response = requests.delete(url, headers=headers, timeout=10)
        if response.status_code == 200:
            return response.json()
        else:
            raise RadarrRequestError(
                "Invalid response received from Radarr: %s" % response.content
            )
    except RequestException as e:
        raise RadarrRequestError("Unable to connect to Radarr at %s. Error: %s" % (url, e))


def request_post_json(url, headers, data):
    """ Makes a POST request and returns the JSON response """
    try:
        response = requests.post(url, headers=headers, data=data, timeout=10)
        if response.status_code == 201:
            return response.json()
        else:
            error_message = None
            try:
                json_response = response.json()
                if len(json_response) > 0 and "errorMessage" in json_response[0]:
                    error_message = json_response[0]["errorMessage"]
            except ValueError:
                # Raised by response.json() if JSON couln't be decoded
                logger.error('Radarr returned non-JSON error result: {}', response.content)

            raise RadarrRequestError(
                "Invalid response received from Radarr: %s" % response.content,
                logger,
                status_code=response.status_code,
                error_message=error_message,
            )

    except RequestException as e:
        raise RadarrRequestError("Unable to connect to Radarr at %s. Error: %s" % (url, e))


def request_put_json(url, headers):
    """ Makes a PUT request and returns the JSON response """
    try:
        response = requests.put(url, headers=headers)
        if response.status_code == 200:
            return response.json()
        else:
            raise RadarrRequestError(
                "Invalid response received from Radarr: %s" % response.content
            )
    except RequestException as e:
        raise RadarrRequestError("Unable to connect to Radarr at %s. Error: %s" % (url, e))


class RadarrAPIService:
    """ Handles all communication with the Radarr REST API """

    def __init__(self, api_key, base_url, port=None):
        self.api_key = api_key
        parsed_base_url = urlparse(base_url)

        if parsed_base_url.port:
            port = int(parsed_base_url.port)

        self.api_url = "%s://%s:%s%s/api/" % (
            parsed_base_url.scheme,
            parsed_base_url.netloc,
            port,
            parsed_base_url.path,
        )

    def get_profiles(self):
        """ Gets all profiles """
        request_url = self.api_url + "profile"
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def get_tags(self):
        """ Gets all tags """
        request_url = self.api_url + "tag"
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def add_tag(self, label):
        """ Adds a tag """
        request_url = self.api_url + "tag"
        headers = self._default_headers()
        data = {"label": label}
        return request_post_json(request_url, headers, json.dumps(data))

    def get_movies(self):
        """ Gets all movies """
        request_url = self.api_url + "movie"
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def get_root_folders(self):
        """ Gets the root folders """
        request_url = self.api_url + "rootfolder"
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def delete_movie(self, movie_id):
        """ Deletes a movie provided by its id """
        request_url = self.api_url + "movie/" + str(movie_id)
        headers = self._default_headers()
        return request_delete_json(request_url, headers)

    def lookup_by_term(self, term):
        """ Returns all movies that matches the search term """
        term = quote(term)
        request_url = self.api_url + "movie/lookup?term=" + term
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def lookup_by_imdb(self, imdb_id):
        """ Returns all movies that matches the imdb id """
        # TODO: make regexp check that imdb_id really is an IMDB_ID
        request_url = self.api_url + "movie/lookup/imdb?imdbId=" + imdb_id
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def lookup_by_tmdb(self, tmdb_id):
        """ Returns all movies that matches the tmdb id """
        tmdb_id = int(tmdb_id)
        request_url = self.api_url + "movie/lookup/tmdb?tmdbId=" + str(tmdb_id)
        headers = self._default_headers()
        return request_get_json(request_url, headers)

    def add_movie(
        self,
        title,
        year,
        quality_profile_id,
        title_slug,
        images,
        tmdb_id,
        root_folder_path,
        monitored=True,
        add_options=None,
        tags=(),
    ):
        """ Adds a movie """
        request_url = self.api_url + "movie"
        headers = self._default_headers()
        data = {
            "title": title,
            "year": year,
            "qualityProfileId": quality_profile_id,
            "titleSlug": title_slug,
            "images": images,
            "tmdbId": tmdb_id,
            "rootFolderPath": root_folder_path,
            "monitored": monitored,
            "tags": tags,
        }

        if add_options:
            data["addOptions"] = add_options

        try:
            json_response = request_post_json(request_url, headers, json.dumps(data))
        except RadarrRequestError as ex:
            spec_ex = spec_exception_from_response_ex(ex)
            if spec_ex:
                raise spec_ex
            else:
                raise

        return json_response

    def _default_headers(self):
        """ Returns a dictionary with default headers """
        return {"X-Api-Key": self.api_key}


# Maps (lowercase) Radarr qualities to flexget
# quality reqirement strings
QUALITIES_MAP = {
    "workprint": "workprint",
    "cam": "cam",
    "telesync": "ts",
    "telecine": "tc",
    "dvdscr": "dvdscr",
    "sdtv": "sdtv",
    "dvd": "dvdrip",  # not completely correct
    "dvd-r": "dvdrip",  # not completely correct
    "webdl-480p": "webdl 480p",
    "bluray-480p": "bluray 480p",
    "bluray-576p": "bluray 576p",
    "hdtv-720p": "hdtv 720p",
    "webdl-720p": "webdl 720p",
    "bluray-720p": "bluray 720p",
    "hdtv-1080p": "hdtv 1080p",
    "webdl-1080p": "webdl 1080p",
    "bluray-1080p": "bluray 1080p",
    "remux-1080p": "remux 1080p",
    "hdtv-2160p": "hdtv 2160p",
    "webdl-2160p": "webdl 2160p",
    "bluray-2160p": "bluray 2160p",
    "remux-2160p": "remux 2160p",
    "br-disk": "remux",  # not completely correct
    "raw-hd": "remux",  # not completely correct
    # No idea of how to map these:
    # 'regional': 'UNKNOWN'
}


def radarr_quality_to_flexget_quality_req(radarr_quality):
    """
    Translates the provided Radarr quality string to a Flexget Requirement instance.
    Returns None if translation is unsuccessful
    """

    # QUALITIES_MAP has its keys in lower case
    radarr_quality = radarr_quality.lower()
    if not radarr_quality in QUALITIES_MAP:
        logger.warning(
            "Did not find a suitible translation for Radarr quality '{}'", radarr_quality
        )
        return None

    flexget_quality_req_string = QUALITIES_MAP[radarr_quality]

    try:
        return Requirements(flexget_quality_req_string)
    except ValueError:
        logger.error(
            'Failed to convert {} into a valid quality requirement', flexget_quality_req_string
        )


def get_flexget_qualities(profile, cutoff_only=False):
    quality_requirements = []
    if cutoff_only:
        name = profile["cutoff"]["name"]
        quality_req = radarr_quality_to_flexget_quality_req(name)
        if quality_req:
            quality_requirements.append(quality_req)
    else:
        for quality in profile["items"]:
            if quality["allowed"]:
                name = quality["quality"]["name"]
                quality_req = radarr_quality_to_flexget_quality_req(name)
                if quality_req:
                    quality_requirements.append(quality_req)

    return quality_requirements


class RadarrSet(MutableSet):
    """ Accesses the Radarr movies using the provided the config """

    def __init__(self, config):
        self.config = config
        self.service = RadarrAPIService(config["api_key"], config["base_url"], config["port"])

        # cache tags
        self._tags = None

        # Class member used for caching the items to avoid
        # unnecessary calls to the Radarr API.
        # We use the self.items property to access it.
        # Just set this to None again later to invalidate the cache.
        self._movie_entries = None

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)

    def get_tag_ids(self, entry):
        tags_ids = []

        if not self._tags:
            self._tags = {t["label"].lower(): t["id"] for t in self.service.get_tags()}

        for tag in self.config.get("tags", []):
            if isinstance(tag, int):
                # Handle tags by id
                if tag not in self._tags.values():
                    logger.error(
                        'Unable to add tag with id {} to entry {} as the tag does not exist in radarr',
                        entry,
                        tag,
                    )
                    continue
                tags_ids.append(tag)
            else:
                # Handle tags by name
                tag = entry.render(tag).lower()
                found = self._tags.get(tag)
                if not found:
                    logger.verbose('Adding missing tag {} to Radarr', tag)
                    found = self.service.add_tag(tag)["id"]
                    self._tags[tag] = found
                tags_ids.append(found)
        return tags_ids

    def discard(self, entry):
        if not entry:
            return

        matching_entry = self._find_matching_entry(entry)
        if matching_entry:
            movie_id = matching_entry["radarr_id"]
            self.service.delete_movie(movie_id)
            logger.verbose('Removed movie {} from Radarr', matching_entry['title'])
            # Clear the cache
            self._movie_entries = None
        else:
            logger.debug('Could not find any matching movie to remove for entry {}', entry)

    def __ior__(self, other):
        for entry in other:
            self.add(entry)

    def __contains__(self, entry):
        if not entry:
            return False

        matching_entry = self._find_matching_entry(entry)

        return matching_entry is not None

    def add(self, entry):
        # The easiest way to add a movie to Radarr is to
        # first use the lookup API. Using that we will get
        # a json response which gives us most of the input
        # we need for the POST request.
        result = self._lookup_movie(entry.get("title"), entry.get("imdb_id"), entry.get("tmdb_id"))

        if result:
            root_folders = self.service.get_root_folders()
            root_folder_path = root_folders[0]["path"]

            try:
                self.service.add_movie(
                    result["title"],
                    result["year"],
                    self.config.get("profile_id"),
                    result["titleSlug"],
                    result["images"],
                    result["tmdbId"],
                    root_folder_path,
                    monitored=self.config.get('monitored', False),
                    tags=self.get_tag_ids(entry),
                )
                logger.verbose('Added movie {} to Radarr list', result['title'])
            except RadarrMovieAlreadyExistsError:
                logger.warning(
                    'Could not add movie {} because it already exists on Radarr', result['title']
                )
            except RadarrRequestError as ex:
                msg = 'The movie add command raised exception: %s' % ex
                logger.error(msg)
                entry.fail(msg)
        else:
            msg = (
                'The lookup for entry %s did not return any results.Can not add the movie in Radarr.'
                % entry
            )
            logger.verbose(msg)
            entry.fail(msg)

    def _from_iterable(self, it):
        # The following implementation is what's done in every other
        # list plugin. Does not makes sense.
        # As I understand it, it aims to be an override to base class Set._from_iterable.
        # However, thats a classmethod and this does not even match its signature
        # https://blog.devzero.com/2013/01/28/how-to-override-a-class-method-in-python/
        return set(it)

    def get(self, entry):
        # Here we must return a matching entry using the provided the "hint" entry
        # In Radarr's case we will check different movie ids and then resort to name/year.
        return self._find_matching_entry(entry)

    @property
    def items(self):
        """ Property that returns all items and only loads them all items when needed """
        if self._movie_entries is None:
            self._movie_entries = self._get_movie_entries()
        return self._movie_entries

    @property
    def tags(self):
        """ Property that returns tag by id """
        tags_ids = []
        if self._tags is None:
            existing = {t["label"].lower(): t["id"] for t in self.service.get_tags()}
            for tag in self.config_tags:
                tag = tag.lower()
                found = existing.get(tag)
                if not found:
                    logger.verbose('Adding missing tag {}} to Radarr', tag)
                    found = self.service.add_tag(tag)["id"]
                tags_ids.append(found)
            self._tags = tags_ids
        return self._tags

    @property
    def immutable(self):
        # Here we return True if it's not possible to modify the list.
        # Could depends on the configuration (self.config).
        # But as long it's only about movies it should be alright
        # for Radarr
        return False

    @property
    def online(self):
        # Radarr is an online service, so yes...
        return True

    def _find_matching_entry(self, entry):
        """
        Finds a movie by first checking against the ids of the
        provided entry, and if none matches, check by title name

        :returns entry or None
        """
        for movie_entry in self.items:
            # First check if any of the id attributes match
            for id_attribute in ["tmdb_id", "imdb_id", "radarr_id"]:
                if id_attribute in entry and id_attribute in movie_entry:
                    if entry[id_attribute] == movie_entry[id_attribute]:
                        # Perfect match!
                        return movie_entry

            # Then we check if the title matches
            movie_name = entry.get("movie_name")
            movie_year = entry.get("movie_year")
            if movie_name:
                if movie_name.lower() == movie_entry["movie_name"].lower():
                    # The name matches. If we also have a year lets check that as well.
                    if movie_year == movie_entry.get("movie_year", object()):
                        # Movie name and year matches
                        return movie_entry
                    else:
                        # The movie had no year present
                        return movie_entry

            # Last resort is just to compare the title straight off
            title = entry.get("title").lower()
            if title == movie_entry["title"].lower():
                return movie_entry

    def _get_movie_entries(self):
        """
        Returns a collection of Entry instances
        that represents the entries in the Radarr
        movie list

        :returns list of entries
        """
        profiles = self.service.get_profiles()
        movies = self.service.get_movies()

        profile_to_requirement_cache = {}
        entries = []
        for movie in movies:
            if self.config.get("only_monitored") and not movie["monitored"]:
                continue

            quality_requirements = []

            # Check if we should add quality requirement
            if self.config.get("include_data"):
                movie_profile_id = movie["profileId"]
                for profile in profiles:
                    profile_id = profile["id"]
                    if profile_id == movie_profile_id:
                        if profile_id not in profile_to_requirement_cache:
                            profile_to_requirement_cache[profile_id] = get_flexget_qualities(
                                profile, self.config["only_use_cutoff_quality"]
                            )

                        quality_requirements = profile_to_requirement_cache[profile_id]
                        break

            entry = Entry(
                title=movie["title"],
                url="",
                radarr_id=movie["id"],
                movie_name=movie["title"],
                movie_year=movie["year"],
            )

            # There seem to be a bug in the Radarr API because sometimes
            # the imdbId is omitted in the response. So we can't be sure
            # it's there
            if "imdbId" in movie:
                entry["imdb_id"] = movie["imdbId"]

            if "tmdbId" in movie:
                entry["tmdb_id"] = movie["tmdbId"]

            if len(quality_requirements) > 0:
                entry["quality_req"] = [str(quality_req) for quality_req in quality_requirements]

            entries.append(entry)

        return entries

    # TODO: this fails
    def _lookup_movie(self, title=None, imdb_id=None, tmdb_id=None):
        """
        Uses Radarr's API to lookup a movie, prioritizing IMDB/TMDB
        ids and as a last resort search for the title
        """
        # If the entry has a IMDB id, use that for lookup
        if imdb_id:
            try:
                result = self.service.lookup_by_imdb(imdb_id)
                # lookup_by_imdb returns an empty dictionary in case no match is found
                if result:
                    return result
            except RadarrRequestError as ex:
                logger.error('Radarr IMDB lookup failed: {}', ex)

        # If the entry has a TMDB id, use that for lookup
        if tmdb_id:
            try:
                result = self.service.lookup_by_tmdb(tmdb_id)
                # lookup_by_tmdb returns an empty dictionary in case no match is found
                if result:
                    return result
            except RadarrRequestError as ex:
                logger.error('Radarr TMDB lookup failed: {}', ex)

        # Could not lookup by id. Try to use the title.
        # However, we can only accept any results if it's
        # one item, otherwise we don't know which to select.
        if title:
            try:
                results = self.service.lookup_by_term(title)
                if len(results) > 1:
                    logger.debug(
                        "Radarr lookup for '{}' returned {:d} results. Using the first result '{}'.",
                        title,
                        len(results),
                        results[0]['title'],
                    )
                    return results[0]
            except RadarrRequestError as ex:
                logger.error('Radarr search term lookup failed: {}', ex)


class RadarrList:
    """ List plugin for Radarr that also works as an input plugin """

    schema = {
        "type": "object",
        "properties": {
            "base_url": {"type": "string"},
            "port": {"type": "number", "default": 80},
            "api_key": {"type": "string"},
            "only_monitored": {"type": "boolean", "default": True},
            "include_data": {"type": "boolean", "default": False},
            "only_use_cutoff_quality": {"type": "boolean", "default": False},
            "monitored": {"type": "boolean", "default": True},
            "profile_id": {"type": "integer", "default": 1},
            "tags": {"type": "array", "items": {'type': ['integer', 'string']}},
        },
        "required": ["api_key", "base_url"],
        "additionalProperties": False,
    }

    @staticmethod
    def get_list(config):
        # Called when used as a list plugin
        return RadarrSet(config)

    def on_task_input(self, task, config):
        # Called when used as an input plugin
        return list(RadarrSet(config))


@event("plugin.register")
def register_plugin():
    plugin.register(RadarrList, "radarr_list", api_ver=2, interfaces=["task", "list"])



import json

import pytest

from flexget.api.app import base_message
from flexget.api.core.database import ObjectsContainer as OC


class TestDatabaseAPI:
    config = 'tasks: {}'

    @pytest.mark.parametrize(
        'operation, plugin_name, status, schema',
        [
            ('cleanup', None, 200, base_message),
            ('vacuum', None, 200, base_message),
            ('plugin_reset', 'bla', 400, base_message),
            ('plugin_reset', 'tvmaze', 200, base_message),
        ],
    )
    def test_database_methods(
        self, operation, plugin_name, status, schema, api_client, schema_match
    ):
        payload = {'operation': operation}
        if plugin_name:
            payload['plugin_name'] = plugin_name

        rsp = api_client.json_post('/database/', data=json.dumps(payload))
        assert rsp.status_code == status, 'Response code is %s' % rsp.status_code
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(schema, data)
        assert not errors

    def test_database_get_plugins(self, api_client, schema_match):
        rsp = api_client.get('/database/plugins/')
        assert rsp.status_code == 200
        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(OC.plugin_list, data)
        assert not errors



# -*- coding: utf-8 -*-
import pytest

from flexget.api.app import base_message
from flexget.components.trakt.api import ObjectsContainer as oc
from flexget.utils import json


@pytest.mark.online
class TestTraktSeriesLookupAPI:
    config = 'tasks: {}'

    def test_trakt_series_lookup_no_params(self, api_client, schema_match):
        # Bad API call
        rsp = api_client.get('/trakt/series/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code

        rsp = api_client.get('/trakt/series/?title=the x-files')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(oc.series_return_object, data)
        assert not errors

        values = {
            'id': 4063,
            'imdb_id': 'tt0106179',
            'language': 'en',
            'title': 'The X-Files',
            'tmdb_id': 4087,
            'tvdb_id': 77398,
            'tvrage_id': 6312,
            'year': 1993,
        }

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_year_param(self, api_client, schema_match):
        values = {
            'id': 235,
            'imdb_id': 'tt0098798',
            'language': 'en',
            'title': 'The Flash',
            'tmdb_id': 236,
            'tvdb_id': 78650,
            'tvrage_id': 5781,
            'year': 1990,
        }

        rsp = api_client.get('/trakt/series/?title=the flash&year=1990')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_trakt_slug_id_param(self, api_client, schema_match):
        values = {'id': 75481, 'title': 'The Flash', 'tvdb_id': 272094, 'year': 1967}

        rsp = api_client.get('/trakt/series/?title=the flash&trakt_slug=the-flash-1967')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))

        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_tmdb_id_param(self, api_client, schema_match):
        values = {
            'id': 60300,
            'imdb_id': 'tt3107288',
            'title': 'The Flash',
            'tmdb_id': 60735,
            'tvdb_id': 279121,
            'tvrage_id': 36939,
            'year': 2014,
        }

        rsp = api_client.get('/trakt/series/?title=the flash&tmdb_id=60735')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_imdb_id_param(self, api_client):
        values = {
            'id': 60300,
            'imdb_id': 'tt3107288',
            'title': 'The Flash',
            'tmdb_id': 60735,
            'tvdb_id': 279121,
            'tvrage_id': 36939,
            'year': 2014,
        }

        rsp = api_client.get('/trakt/series/?title=the flash&imdb_id=tt3107288')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_tvdb_id_param(self, api_client, schema_match):
        values = {
            'id': 60300,
            'imdb_id': 'tt3107288',
            'title': 'The Flash',
            'tmdb_id': 60735,
            'tvdb_id': 279121,
            'tvrage_id': 36939,
            'year': 2014,
        }

        rsp = api_client.get('/trakt/series/?title=the flash&tvdb_id=279121')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_tvrage_id_param(self, api_client, schema_match):
        values = {
            'id': 60300,
            'imdb_id': 'tt3107288',
            'title': 'The Flash',
            'tmdb_id': 60735,
            'tvdb_id': 279121,
            'tvrage_id': 36939,
            'year': 2014,
        }

        rsp = api_client.get('/trakt/series/?title=the flash&tvrage_id=36939')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_trakt_id_param(self, api_client, schema_match):
        values = {'id': 75481, 'title': 'The Flash', 'year': 1967}

        rsp = api_client.get('/trakt/series/?title=the flash&trakt_id=75481')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_series_lookup_with_actors_param(self, api_client, schema_match):
        rsp = api_client.get('/trakt/series/?title=the x-files&include_actors=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        assert 'actors' in data
        assert len(data['actors']) > 0

    def test_trakt_series_lookup_with_translations_param(self, api_client, schema_match):
        rsp = api_client.get('/trakt/series/?title=game of thrones&include_translations=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.series_return_object, data)
        assert not errors

        assert 'translations' in data

    def test_trakt_series_lookup_error(self, api_client, schema_match):
        rsp = api_client.get('/trakt/series/?title=sdfgsdfgsdfgsdfgsdfg')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(base_message, data)
        assert not errors


@pytest.mark.online
class TestTraktMovieLookupAPI:
    config = 'tasks: {}'

    def test_trakt_movies_lookup_no_params(self, api_client, schema_match):
        # Bad API call
        rsp = api_client.get('/trakt/movies/')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code

        rsp = api_client.get('/trakt/movies/?title=the matrix')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_return_object, data)
        assert not errors

        values = {
            'id': 481,
            'title': 'The Matrix',
            'year': 1999,
            'tmdb_id': 603,
            'imdb_id': 'tt0133093',
        }
        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_movies_lookup_year_param(self, api_client, schema_match):
        rsp = api_client.get('/trakt/movies/?title=the matrix&year=2003')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_return_object, data)
        assert not errors

        values = {
            'id': 483,
            'title': 'The Matrix Revolutions',
            'year': 2003,
            'tmdb_id': 605,
            'imdb_id': 'tt0242653',
        }
        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_movies_lookup_slug_param(self, api_client, schema_match):
        rsp = api_client.get('/trakt/movies/?title=the matrix&trakt_slug=the-matrix-reloaded-2003')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_return_object, data)
        assert not errors

        values = {
            'id': 482,
            'title': 'The Matrix Reloaded',
            'year': 2003,
            'tmdb_id': 604,
            'imdb_id': 'tt0234215',
        }
        for field, value in values.items():
            assert data.get(field) == value

    def test_trakt_movies_lookup_actors_params(self, api_client, schema_match):
        rsp = api_client.get('/trakt/movies/?title=the matrix&include_actors=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_return_object, data)
        assert not errors

        values = {
            'id': 481,
            'title': 'The Matrix',
            'year': 1999,
            'tmdb_id': 603,
            'imdb_id': 'tt0133093',
        }
        for field, value in values.items():
            assert data.get(field) == value

        assert 'actors' in data
        assert len(data['actors']) > 0

    def test_trakt_movies_lookup_translations_params(self, api_client, schema_match):
        rsp = api_client.get('/trakt/movies/?title=the matrix&include_translations=true')
        assert rsp.status_code == 200, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(oc.movie_return_object, data)
        assert not errors

        values = {
            'id': 481,
            'title': 'The Matrix',
            'year': 1999,
            'tmdb_id': 603,
            'imdb_id': 'tt0133093',
        }
        for field, value in values.items():
            assert data.get(field) == value

        assert 'translations' in data
        assert len(data['translations']) > 0

    def test_trakt_movies_lookup_error(self, api_client, schema_match):
        rsp = api_client.get('/trakt/movies/?title=sdfgsdfgsdfgsdfgsdfg')
        assert rsp.status_code == 404, 'Response code is %s' % rsp.status_code

        data = json.loads(rsp.get_data(as_text=True))
        errors = schema_match(base_message, data)
        assert not errors



import itertools

from loguru import logger

from flexget import plugin
from flexget.event import event

logger = logger.bind(name='limit')


class PluginLimit:
    """
    Limits the number of entries an input plugin can produce.
    """

    schema = {
        'type': 'object',
        'properties': {
            'amount': {'type': 'integer', 'minimum': -1},
            'from': {
                'allOf': [
                    {'$ref': '/schema/plugins?phase=input'},
                    {
                        'maxProperties': 1,
                        'error_maxProperties': 'Plugin options within limit plugin must be indented 2 more spaces than '
                        'the first letter of the plugin name.',
                        'minProperties': 1,
                    },
                ]
            },
        },
        'required': ['amount', 'from'],
        'additionalProperties': False,
    }

    def on_task_input(self, task, config):
        for input_name, input_config in config['from'].items():
            input = plugin.get_plugin_by_name(input_name)
            method = input.phase_handlers['input']
            result = method(task, input_config)
            # A 0 or -1 limit means don't limit.
            if config['amount'] < 1:
                return result
            return itertools.islice(result, config['amount'])


@event('plugin.register')
def register_plugin():
    plugin.register(PluginLimit, 'limit', api_ver=2)



from datetime import timedelta

from loguru import logger
from sqlalchemy import desc, func

from flexget import plugin
from flexget.event import event
from flexget.manager import Session
from flexget.utils.tools import multiply_timedelta

from . import db

logger = logger.bind(name='est_series_internal')


class EstimatesSeriesInternal:
    @plugin.priority(0)  # Should always be last priority
    def estimate(self, entry):
        if not all(field in entry for field in ['series_name', 'series_season', 'series_episode']):
            return
        with Session() as session:
            series = (
                session.query(db.Series).filter(db.Series.name == entry['series_name']).first()
            )
            if not series:
                return
            episodes = (
                session.query(db.Episode)
                .join(db.Episode.series)
                .filter(db.Episode.season != None)
                .filter(db.Series.id == series.id)
                .filter(db.Episode.season == func.max(db.Episode.season).select())
                .order_by(desc(db.Episode.number))
                .limit(2)
                .all()
            )

            if len(episodes) < 2:
                return
            # If last two eps were not contiguous, don't guess
            if episodes[0].number != episodes[1].number + 1:
                return
            # If first_seen in None, return
            if episodes[0].first_seen is None or episodes[1].first_seen is None:
                return
            last_diff = episodes[0].first_seen - episodes[1].first_seen
            # If last eps were grabbed close together, we might be catching up, don't guess
            # Or, if last eps were too far apart, don't guess
            # TODO: What range?
            if last_diff < timedelta(days=2) or last_diff > timedelta(days=10):
                return
            # Estimate next season somewhat more than a normal episode break
            if entry['series_season'] > episodes[0].season:
                # TODO: How big should this be?
                return episodes[0].first_seen + multiply_timedelta(last_diff, 2)
            # Estimate next episode comes out about same length as last ep span, with a little leeway
            return episodes[0].first_seen + multiply_timedelta(last_diff, 0.9)


@event('plugin.register')
def register_plugin():
    plugin.register(
        EstimatesSeriesInternal, 'est_series_internal', interfaces=['estimate_release'], api_ver=2
    )



from loguru import logger

from flexget import plugin
from flexget.event import event
from flexget.plugin import PluginError

logger = logger.bind(name='list_clear')


class ListClear:
    schema = {
        'type': 'object',
        'properties': {
            'what': {
                'type': 'array',
                'items': {
                    'allOf': [
                        {'$ref': '/schema/plugins?interface=list'},
                        {
                            'maxProperties': 1,
                            'error_maxProperties': 'Plugin options within list_clear plugin must be indented '
                            '2 more spaces than the first letter of the plugin name.',
                            'minProperties': 1,
                        },
                    ]
                },
            },
            'phase': {'type': 'string', 'enum': plugin.task_phases, 'default': 'start'},
        },
        'required': ['what'],
    }

    def __getattr__(self, phase):
        # enable plugin in regular task phases
        if phase.replace('on_task_', '') in plugin.task_phases:
            return self.clear

    @plugin.priority(plugin.PRIORITY_FIRST)
    def clear(self, task, config):
        for item in config['what']:
            for plugin_name, plugin_config in item.items():
                try:
                    thelist = plugin.get(plugin_name, self).get_list(plugin_config)
                except AttributeError:
                    raise PluginError('Plugin %s does not support list interface' % plugin_name)
                if thelist.immutable:
                    raise plugin.PluginError(thelist.immutable)
                if config['phase'] == task.current_phase:
                    if task.manager.options.test and thelist.online:
                        logger.info(
                            'would have cleared all items from {} - {}', plugin_name, plugin_config
                        )
                        continue
                    logger.verbose('clearing all items from {} - {}', plugin_name, plugin_config)
                    thelist.clear()


@event('plugin.register')
def register_plugin():
    plugin.register(ListClear, 'list_clear', api_ver=2)



from loguru import logger

from flexget import plugin
from flexget.config_schema import one_or_more
from flexget.event import event
from flexget.utils import qualities

logger = logger.bind(name='quality')


class FilterQuality:
    """
    Rejects all entries that don't have one of the specified qualities

    Example::

      quality:
        - hdtv
    """

    schema = one_or_more({'type': 'string', 'format': 'quality_requirements'})

    # Run before series and imdb plugins, so correct qualities are chosen
    @plugin.priority(175)
    def on_task_filter(self, task, config):
        if not isinstance(config, list):
            config = [config]
        reqs = [qualities.Requirements(req) for req in config]
        for entry in task.entries:
            if entry.get('quality') is None:
                entry.reject('Entry doesn\'t have a quality')
                continue
            if not any(req.allows(entry['quality']) for req in reqs):
                text_reqs = ', '.join(f'`{req}`' for req in reqs)
                entry.reject(
                    f'`{entry["quality"]}` does not match any of quality requirements: {text_reqs}'
                )


@event('plugin.register')
def register_plugin():
    plugin.register(FilterQuality, 'quality', api_ver=2)
